# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from asreview.project import ProjectNotFoundError
from asreview.project import is_project


def read_data(project, use_cache=True, save_cache=True):
    """Get ASReviewData object from file.

    Parameters
    ----------
    project_path: str, iterable
        The project identifier.
    use_cache: bool
        Use the pickle file if available.
    save_cache: bool
        Save the file to a pickle file if not available.

    Returns
    -------
    ASReviewData:
        The data object for internal use in ASReview.

    """

    logging.warning(
        "read_data is deprecated, use ASReviewProject.read_data instead. "
        "read_data will be removed in version 2.0."
    )

    if not is_project(project.project_path):
        raise ProjectNotFoundError()

    return project.read_data(use_cache=use_cache, save_cache=save_cache)



import time

from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from asreview.webapp.authentication.models import Project
from asreview.webapp.authentication.models import User


def setup_database_session(uri):
    Session = sessionmaker()
    engine = create_engine(uri)
    Session.configure(bind=engine)
    return Session()


def clean_database(session):
    session.query(Project).delete()
    session.query(User).delete()
    session.commit()


def browse_to_page(driver, page):
    driver.get(page)
    assert driver.current_url == page


def click_element(driver, selector, wait_secs=60):
    """Waits until a clickable element is clickable and clicks
    on it. When <selector> is a String, a CSS selector is
    assumed."""
    if isinstance(selector, str):
        # assume a CSS selector if selector is a string
        selector = (By.CSS_SELECTOR, selector)
    # wait until clickable
    WebDriverWait(driver, wait_secs) \
        .until(EC.element_to_be_clickable(selector))
    # find the element
    element = driver.find_element(*selector)
    # click
    element.click()
    return element


def select_from_dropdown(driver, parent, data_value):
    click_element(driver, parent)
    element = (By.XPATH, f"//li[@data-value=\"{data_value}\"]")
    click_element(driver, element)
    WebDriverWait(driver, 60) \
        .until(EC.invisibility_of_element_located(element))


# TODO APPLY THIS FUNCTION
def fill_text_field_by_id(driver, field_id, value):
    WebDriverWait(driver, 60) \
        .until(EC.presence_of_element_located(
            (By.CSS_SELECTOR, f"input#{field_id}")))

    input_field = driver.find_element(
        By.CSS_SELECTOR,
        f"input#{field_id}"
    )
    input_field.clear()
    input_field.send_keys(value)


def create_account(driver, base_url, account_data):
    # browse to signup page
    page = base_url + "/signup"
    browse_to_page(driver, page)

    # link account data to form fields
    form_field_values = [
        ("email", account_data["email"]),
        ("name", account_data["name"]),
        ("affiliation", account_data["affiliation"]),
        ("password", account_data["password"]),
        ("confirmPassword", account_data["password"]),
    ]

    # make sure we got the input fields ready
    WebDriverWait(driver, 60) \
        .until(EC.presence_of_element_located(
            (By.CSS_SELECTOR, f"input#{form_field_values[0][0]}")))

    # enter form data
    for field_id, value in form_field_values:
        driver.find_element(
            By.CSS_SELECTOR,
            f"input#{field_id}"
        ).send_keys(value)

    # wait until create button is clickable and create profile
    click_element(driver, "button#create-profile")


def sign_in(driver, base_url, account_data):
    # browse to signin page
    page = base_url + "/signin"
    browse_to_page(driver, page)

    # compile form data
    form_field_values = [
        ("email", account_data["email"]),
        ("password", account_data["password"]),
    ]

    # make sure we got the input fields ready
    WebDriverWait(driver, 60) \
        .until(EC.presence_of_element_located(
            (By.CSS_SELECTOR, f"input#{form_field_values[0][0]}")))

    # enter form data
    for field_id, value in form_field_values:
        driver.find_element(
            By.CSS_SELECTOR,
            f"input#{field_id}"
        ).send_keys(value)

    # click on signin button (it must be visible)
    click_element(driver, "button#sign-in")


def sign_out(driver):
    click_element(driver, "button#profile-popper")
    click_element(driver, "li#signout")


def page_contains_text(driver, text, css_selector=None):
    if css_selector is None:
        css_selector = "body"
    # find the selector
    all_text = driver.find_element(By.CSS_SELECTOR, css_selector).text
    return text in all_text


def label_abstract(driver, label, reading_time=0):
    # sleep to simulate reading time
    time.sleep(reading_time)
    # make sure we're dealing with lowercase characters
    label = label.lower().strip()
    # check if label is correct
    assert label in ["irrelevant", "relevant"]
    # click appropriate button
    click_element(driver, f"button#{label}")


def _label_prior_knowledge_abstract(driver, label=None, reading_time=0):
    # Use prior label if label is None
    if label is None:
        # find first element that has class labeled-as
        label = driver.find_element(By.CSS_SELECTOR, "div.labeled-as").text
    # proceed with actual labeling
    label_abstract(driver, label, reading_time)


def _label_random_prior_knowledge(driver, project_data, reading_time):
    # click on randomly adding prior knowledge
    click_element(
        driver,
        (By.XPATH, "//span[text()=\"Random\"]")
    )

    # add random prior knowledge
    for label in project_data["dataset"]["prior_knowledge"]:
        _label_prior_knowledge_abstract(driver, label, reading_time)


def _label_searched_prior_knowledge(driver, project_data, reading_time):
    # make the Search choice
    click_element(
        driver,
        (By.XPATH, "//span[text()=\"Search\"]")
    )

    # Work the expected search terms, send the strings and wait
    # for the result before we can label.
    for search_term, label in project_data["dataset"]["prior_knowledge"]:
        # fill out search input
        driver.find_element(
            By.CSS_SELECTOR,
            "input#search-input"
        ).send_keys(search_term)
        # click search button
        click_element(driver, "button#search")

        # wait until we have a result
        WebDriverWait(driver, 60) \
            .until(EC.presence_of_element_located(
                (By.CSS_SELECTOR, "div.search-result")))

        # label first abstract
        _label_prior_knowledge_abstract(driver, label, reading_time)

        # clear search input
        driver.find_element(
            By.CSS_SELECTOR,
            "input#search-input"
        ).clear()


def create_project(driver, base_url, project_data, reading_time=0):
    # browse to signin page
    page = base_url + "/projects"
    browse_to_page(driver, page)

    # click on the create button when it's available
    click_element(driver, "button#create-project")

    # PAGE 1 OF MODAL
    # starting with the mode
    select_from_dropdown(driver, "div#mode-select", project_data['mode'])

    form_field_values = [
        ("input#project-title", project_data["title"]),
        ("input#project-author", project_data["author"]),
        ("textarea#project-description", project_data["description"])
    ]

    # enter form data
    for field_selector, value in form_field_values:
        driver.find_element(
            By.CSS_SELECTOR,
            field_selector,
        ).send_keys(value)

    # click on next
    click_element(driver, "button#next")

    # PAGE 2, DATASET AND PRIOR KNOWLEDGE
    # adding a dataset
    click_element(driver, "button#add-dataset")

    # choose the dataset
    dataset_button = click_element(
        driver,
        (By.XPATH, f"//p[text()=\"{project_data['dataset']['label']}\"]")
    )
    # find the button in the parent and click it (it's clickable)
    dataset_button.find_element(By.XPATH, "../../../..//button").click()

    # Adding prior knowledge
    click_element(driver, "button#add-prior-knowledge")

    if project_data["dataset"]["prior_knowledge_method"] == "Random":
        _label_random_prior_knowledge(driver, project_data, reading_time)
    else:
        _label_searched_prior_knowledge(driver, project_data, reading_time)

    # close page 2
    click_element(
        driver,
        (By.XPATH, "//button[text()='Close']")
    )

    # click on next
    click_element(driver, "button#next")

    # PAGE 3, MODEL
    # adding feature extraction mode
    feature_extr_type = project_data["model"]["feature_extraction"]
    select_from_dropdown(
        driver,
        "div#select-feature-extraction",
        feature_extr_type
    )

    # classifier
    classifier = project_data["model"]["classifier"]
    select_from_dropdown(
        driver,
        "div#select-classifier",
        classifier
    )

    # query stratgey
    query_strategy = project_data["model"]["query_strategy"]
    select_from_dropdown(
        driver,
        "div#select-query-strategy",
        query_strategy
    )

    # balance strategy
    balance_strategy = project_data["model"]["balance_strategy"]
    select_from_dropdown(
        driver,
        "div#select-balance-strategy",
        balance_strategy
    )

    # click on next
    click_element(driver, "button#next")

    # Wait until model is trained and so we can start
    click_element(driver, "button#start-reviewing", wait_secs=300)



from pathlib import Path

import pytest
from sqlalchemy.exc import IntegrityError

import asreview.webapp.tests.utils.crud as crud
from asreview.utils import asreview_path
from asreview.webapp import DB
from asreview.webapp.authentication.models import Project

# NOTE: projects are created from a user account

# #############
# CREATE
# #############


# test uniqueness of project id
def test_uniqueness_of_project_id(user):
    project_id = "my-project"
    crud.create_project(DB, user, Project(project_id=project_id))
    assert crud.count_projects() == 1
    with pytest.raises(IntegrityError):
        crud.create_project(DB, user, Project(project_id=project_id))


# insert a project successfully
def test_inserting_project(user):
    project_id = "my-project"
    project = Project(project_id=project_id)
    crud.create_project(DB, user, project)
    assert crud.count_projects() == 1
    # get project
    project = crud.last_project()
    assert project.project_id == project_id
    assert project.owner_id == user.id


# #############
# DELETE
# #############


# deleting a project won't delete its owner
def test_not_delete_user_after_deletion_project(user):
    crud.create_project(DB, user, Project(project_id="project"))
    assert crud.count_users() == 1
    assert crud.count_projects() == 1
    # get project
    project = crud.last_project()
    # delete
    DB.session.delete(project)
    DB.session.commit()
    assert crud.count_users() == 1
    assert crud.count_projects() == 0


# deleting a project will remove invitations
def test_project_removal_invitations(user):
    project = crud.create_project(DB, user, Project(project_id="my-project"))
    user2 = crud.create_user(DB, user=2)
    assert crud.count_users() == 2
    assert crud.count_projects() == 1
    assert crud.count_invitations() == 0
    # invite
    project.pending_invitations.append(user2)
    DB.session.commit()
    assert crud.count_invitations() == 1
    DB.session.delete(project)
    DB.session.commit()
    assert crud.count_projects() == 0
    assert crud.count_invitations() == 0


# deleting a project will remove collaboration links
def test_project_removal_collaborations(user):
    project = crud.create_project(DB, user, Project(project_id="my-project"))
    user2 = crud.create_user(DB, user=2)
    assert crud.count_users() == 2
    assert crud.count_projects() == 1
    assert crud.count_collaborations() == 0
    # invite
    project.collaborators.append(user2)
    DB.session.commit()
    assert crud.count_collaborations() == 1
    DB.session.delete(project)
    DB.session.commit()
    assert crud.count_projects() == 0
    assert crud.count_collaborations() == 0


# #############
# PROPERTIES
# #############


# test getting a user from project
def test_getting_user_from_project(user):
    project_id = "my-project"
    project = Project(project_id=project_id)
    crud.create_project(DB, user, project)
    assert crud.count_projects() == 1
    # get project
    project = crud.last_project()
    assert project.owner == user


# test project_folder
def test_project_folder(user):
    project_id = "my-project"
    crud.create_project(DB, user, Project(project_id=project_id))
    assert crud.count_projects() == 1
    # get project
    project = crud.last_project()
    assert project.folder == project_id


# test project_path
def test_project_path(user):
    project_id = "my-project"
    crud.create_project(DB, user, Project(project_id=project_id))
    assert crud.count_projects() == 1
    # get project
    project = crud.last_project()
    assert project.project_path == Path(asreview_path() / project_id)


# test pending invites
def test_pending_invites(user):
    project = crud.create_project(DB, user, Project(project_id="my-project"))
    user2 = crud.create_user(DB, user=2)
    assert crud.count_users() == 2
    assert crud.count_projects() == 1
    # invite
    project.pending_invitations.append(user2)
    DB.session.commit()
    # fresh object
    project = crud.last_project()
    # asserts
    assert user2 in project.pending_invitations


# test collaboration
def test_collaboration(user):
    project = crud.create_project(DB, user, Project(project_id="my-project"))
    user2 = crud.create_user(DB, user=2)
    assert crud.count_users() == 2
    assert crud.count_projects() == 1
    assert crud.count_collaborations() == 0
    # invite
    project.collaborators.append(user2)
    DB.session.commit()
    assert crud.count_collaborations() == 1
    # start with a fresh object
    project = crud.last_project()
    assert user2 in project.collaborators



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["BaseTrainClassifier"]

from asreview.models.base import BaseModel


class BaseTrainClassifier(BaseModel):
    """
    Base model, abstract class to be implemented by derived ones.

    All the non-abstract methods are okay if they are not implemented.
    All functions dealing with hyperparameters can be ignore if you don't
    use hyperopt for hyperparameter tuning.
    There is a distinction between model parameters, which are needed during
    model creation and fit parameters, which are used during the fitting
    process. Fit parameters can be distinct from fit_kwargs (which are passed
    to the fit function).
    """

    name = "base-train"

    def __init__(self):
        self._model = None

    def fit(self, X, y):
        """Fit the model to the data.

        Arguments
        ---------
        X: numpy.ndarray
            Feature matrix to fit.
        y: numpy.ndarray
            Labels for supervised learning.
        """
        return self._model.fit(X, y)

    def predict_proba(self, X):
        """Get the inclusion probability for each sample.

        Arguments
        ---------
        X: numpy.ndarray
            Feature matrix to predict.

        Returns
        -------
        numpy.ndarray
            Array with the probabilities for each class, with two
            columns (class 0, and class 1) and the number of samples rows.
        """
        return self._model.predict_proba(X)

    def full_hyper_space(self):
        """Get a hyperparameter space to use with hyperopt.

        Returns
        -------
        dict, dict
            Parameter space. Parameter choices; in case of hyperparameters
            with a list of choices, store the choices there.
        """
        return {}, {}



import pytest
import sqlalchemy
from sqlalchemy import create_engine

from asreview.utils import asreview_path


def get_db_path():
    return asreview_path() / "asreview.test.sqlite"


# checks if asreview path does not contain a database if app
# is unauthenticated
def test_database_is_not_created_if_unauth_app(unauth_app):
    assert asreview_path().exists()
    assert get_db_path().exists() is False


# checks is asreview path contains database if app is
# authenticated
def test_database_exists_after_starting_auth_app(auth_app):
    assert asreview_path().exists()
    assert get_db_path().exists()


# checks if all tables were created
@pytest.mark.parametrize(
    "table", ["collaboration_invitations", "collaborations", "projects", "users"]
)
def test_if_db_table_exists(auth_app, table):
    engine = create_engine(f"sqlite:///{get_db_path()}")
    table_names = sqlalchemy.inspect(engine).get_table_names()
    assert table in table_names



from inspect import getfullargspec

import pytest

import asreview.webapp.tests.utils.api_utils as au

# NOTE: user 1 is signed in and has a single project, invites
# other users who accept and reject


# Test sending an invitation
def test_user1_sends_invitation(setup_auth):
    client, _, user2, _, project = setup_auth
    # invite
    status_code, resp_data = au.invite(client, project, user2)
    assert status_code == 200
    assert resp_data["message"] == 'User "user2@asreview.nl" invited.'


# Testing listing invitations
def test_user2_list_invitations(setup_auth):
    client, user1, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # get all invitations
    status_code, resp_data = au.list_invitations(client)
    invitations = resp_data["invited_for_projects"]
    assert status_code == 200
    assert len(invitations) == 1
    assert invitations[0]["project_id"] == project.project_id
    assert invitations[0]["owner_id"] == user1.id


# Testing accepting an invitation
def test_user2_accept_invitation(setup_auth):
    client, _, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # accept invitation
    status_code, resp_data = au.accept_invitation(client, project)
    assert status_code == 200
    assert resp_data["message"] == "User accepted invitation for project."


# Test rejecting invitation
def test_user2_rejects_invitation(setup_auth):
    client, _, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # reject invitation
    status_code, resp_data = au.reject_invitation(client, project)
    assert status_code == 200
    assert resp_data["message"] == "User rejected invitation for project."


# Test owner removes invitation
def test_owner_deletes_invitation(setup_auth):
    client, _, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # remove invitation
    status_code, resp_data = au.delete_invitation(client, project, user2)
    assert status_code == 200
    assert resp_data["message"] == "Owner deleted invitation."


# Test owner views collaboration team
def test_view_collaboration_team_with_pending_invitation(setup_auth):
    client, _, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # checks team
    status_code, resp_data = au.list_collaborators(client, project)
    assert status_code == 200
    assert resp_data["collaborators"] == []
    assert resp_data["invitations"] == [user2.id]


# Test owner views collaboration team
def test_view_collaboration_team_with_accepted_invitation(setup_auth):
    client, user1, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # accept invitation and signs out
    au.accept_invitation(client, project)
    au.signout_user(client)
    # user 1 signs up
    au.signin_user(client, user1)
    # checks team
    status_code, resp_data = au.list_collaborators(client, project)
    assert status_code == 200
    assert resp_data["collaborators"] == [user2.id]
    assert resp_data["invitations"] == []


# Test owner removes collaboration
def test_owner_deletes_collaboration(setup_auth):
    client, user1, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # accept invitation and signs out
    au.accept_invitation(client, project)
    au.signout_user(client)
    # user 1 signs up
    au.signin_user(client, user1)
    # remove from team
    status_code, resp_data = au.delete_collaboration(client, project, user2)
    assert status_code == 200
    assert resp_data["message"] == "Collaborator removed from project."


# Test collaborator withdraws from collaboration
def test_collaborator_withdrawal(setup_auth):
    client, _, user2, _, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # accept invitation and signs out
    au.accept_invitation(client, project)
    # withdrawal
    status_code, resp_data = au.delete_collaboration(client, project, user2)
    assert status_code == 200
    assert resp_data["message"] == "Collaborator removed from project."


# ###################
# TEST LOGIN REQUIRED
# ###################


@pytest.mark.parametrize(
    "api_call",
    [
        au.invite,
        au.list_invitations,
        au.list_collaborators,
        au.accept_invitation,
        au.reject_invitation,
        au.delete_invitation,
        au.delete_collaboration,
    ],
)
# Test login required for all api routes
def test_login_required(setup_auth, api_call):
    client, _, user2, _, project = setup_auth
    au.signout_user(client)
    number_of_params = len(getfullargspec(api_call).args)
    if number_of_params == 1:
        status_code, resp_data = api_call(client)
    elif number_of_params == 2:
        status_code, resp_data = api_call(client, project)
    elif number_of_params == 3:
        status_code, resp_data = api_call(client, project, user2)
    # all calls must return a 401:
    assert status_code == 401


# ###################
# TEST NO PERMISSION
# ###################


# Test user3 can't see invite from user 1 to user 2
def test_user3_cant_see_other_invites(setup_auth):
    client, _, user2, user3, project = setup_auth
    # invite to make sure we have an invitation (user1 is signed in)
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 3 (not invited)
    au.signin_user(client, user3)
    # get all invitations
    status_code, resp_data = au.list_invitations(client)
    assert status_code == 200
    assert resp_data["invited_for_projects"] == []


# Test user3 can't accept invite to user 2
def test_user3_cant_reject_invite_of_user_2(setup_auth):
    client, _, user2, user3, project = setup_auth
    # invite to make sure we have an invitation (user1 is signed in)
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 3 (not invited)
    au.signin_user(client, user3)
    status_code, resp_data = au.accept_invitation(client, project)
    assert status_code == 404
    assert resp_data["message"] == "Request can not made by current user."


# Test user3 can't reject invite to user 2
def test_user3_cant_accept_invite_of_user_2(setup_auth):
    client, _, user2, user3, project = setup_auth
    # invite to make sure we have an invitation (user1 is signed in)
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 3 (not invited)
    au.signin_user(client, user3)
    status_code, resp_data = au.reject_invitation(client, project)
    assert status_code == 404
    assert resp_data["message"] == "Request can not made by current user."


# Test user3 can't delete invitation
def test_user3_cant_delete_invitation(setup_auth):
    client, _, user2, user3, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 3 (not invited)
    au.signin_user(client, user3)
    # remove invitation
    status_code, resp_data = au.delete_invitation(client, project, user2)
    assert status_code == 404
    assert resp_data["message"] == "Request can not made by current user."


# Test user3 can't see collaboration team of user 1
def test_user3_cant_see_collaboration_team(setup_auth):
    client, _, user2, user3, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 3 (not invited)
    au.signin_user(client, user3)
    # check team
    status_code, resp_data = au.list_collaborators(client, project)
    assert status_code == 404
    assert resp_data["message"] == "Request can not made by current user."


# Test user3 can't remove collaboration
def test_user3_cant_delete_collaboration(setup_auth):
    client, _, user2, user3, project = setup_auth
    # invite
    au.invite(client, project, user2)
    # signout user 1
    au.signout_user(client)
    # signin user 2
    au.signin_user(client, user2)
    # accept invitation and signs out
    au.accept_invitation(client, project)
    au.signout_user(client)
    # user 3 signs up
    au.signin_user(client, user3)
    # remove from team
    status_code, resp_data = au.delete_collaboration(client, project, user2)
    assert status_code == 404
    assert resp_data["message"] == "Request can not made by current user."



import numpy as np
from pytest import mark

from asreview.models.classifiers import get_classifier
from asreview.models.query import get_query_model
from asreview.models.query import list_query_strategies


@mark.parametrize(
    "query_strategy",
    [
        "max",
        "random",
        "max_random",
        "uncertainty",
        "max_uncertainty",
        "cluster",
    ],
)
@mark.parametrize("n_instances", [0, 1, 5, 50])
@mark.parametrize("n_train", [0, 1, 5, 50])
def test_query(query_strategy, n_instances, n_train):
    n_features = 50
    n_sample = 100
    classifier = get_classifier("rf")

    query_model = get_query_model(query_strategy)
    X = np.random.rand(n_sample, n_features)

    y = np.concatenate((np.zeros(n_sample // 2), np.ones(n_sample // 2)), axis=0)
    order = np.random.permutation(n_sample)
    X = X[order]
    y = y[order]

    classifier.fit(X, y)

    assert isinstance(query_model.param, dict)
    assert query_model.name == query_strategy

    query_idx = query_model.query(X, classifier, n_instances=n_instances)
    assert len(query_idx) == n_instances
    assert len(query_idx) == len(np.unique(query_idx))

    query_idx, relevance_scores = query_model.query(X, classifier,
                                                    return_classifier_scores=True)
    assert len(query_idx) == X.shape[0]
    if relevance_scores is not None:
        assert relevance_scores.shape == (X.shape[0], 2)


def test_query_general():
    assert len(list_query_strategies()) >= 4



import asreview.webapp.tests.utils.config_parser as cp
from asreview.webapp.authentication.models import Collaboration
from asreview.webapp.authentication.models import CollaborationInvitation
from asreview.webapp.authentication.models import Project
from asreview.webapp.authentication.models import User


def create_user(DB, user=1):
    if isinstance(user, int):
        user = cp.get_user(user)
    try:
        DB.session.add(user)
        DB.session.commit()
        user = User.query.order_by(User.id.desc()).first()
    except Exception as exception:
        user = False
        DB.session.rollback()
        DB.session.flush()
        raise exception
    return user


def get_user_by_id(id):
    return User.query.filter_by(id=id).one()


def get_user_by_identifier(id):
    return User.query.filter_by(identifier=id).one()


def list_users():
    return User.query.all()


def count_users():
    return len(User.query.with_entities(User.id).all())


def update_user(DB, user, attribute, value):
    user = get_user_by_identifier(user.identifier)
    setattr(user, attribute, value)
    DB.session.commit()
    return user


def last_user():
    return User.query.order_by(User.id.desc()).first()


def delete_users(DB):
    DB.session.query(User).delete()
    DB.session.commit()


def delete_collaborations(DB):
    DB.session.query(Collaboration).delete()
    DB.session.commit()


def delete_invitations(DB):
    DB.session.query(CollaborationInvitation).delete()
    DB.session.commit()


def delete_projects(DB):
    DB.session.query(Project).delete()
    DB.session.commit()


def delete_everything(DB):
    DB.drop_all()


def create_project(DB, user, project):
    try:
        user.projects.append(project)
        DB.session.commit()
        id = project.project_id
        project = Project.query.filter_by(project_id=id).one()
    except Exception as exception:
        project = False
        DB.session.rollback()
        DB.session.flush()
        raise exception
    return project


def get_project_by_project_id(id):
    return Project.query.filter_by(project_id=id).one()


def list_projects():
    return Project.query.all()


def count_projects():
    return len(Project.query.with_entities(Project.id).all())


def last_project():
    return Project.query.order_by(Project.id.desc()).first()


def create_invitation(DB, project, user):
    try:
        inv = CollaborationInvitation(project_id=project.id, user_id=user.id)
        DB.session.add(inv)
        DB.session.commit()
    except Exception as exception:
        DB.session.rollback()
        DB.session.flush()
        raise exception


def list_invitations():
    return CollaborationInvitation.query.all()


def last_invitation():
    return CollaborationInvitation.query.order_by(
        CollaborationInvitation.id.desc()
    ).first()


def count_invitations():
    return len(
        CollaborationInvitation.query.with_entities(CollaborationInvitation.id).all()
    )


def create_collaboration(DB, project, user):
    try:
        coll = Collaboration(project_id=project.id, user_id=user.id)
        DB.session.add(coll)
        DB.session.commit()
    except Exception as exception:
        DB.session.rollback()
        DB.session.flush()
        raise exception


def list_collaborations():
    return Collaboration.query.all()


def last_collaboration():
    return Collaboration.query.order_by(Collaboration.id.desc()).first()


def count_collaborations():
    return len(Collaboration.query.with_entities(Collaboration.id).all())


def create_user1_with_2_projects(DB):
    user = create_user(DB)
    project_ids = ["project-1", "project-2"]
    projects = [Project(project_id=p) for p in project_ids]
    user.projects = projects
    DB.session.commit()
    return user, projects



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["ExcelReader"]

import pandas as pd

from asreview.config import COLUMN_DEFINITIONS
from asreview.io.utils import _standardize_dataframe


class ExcelReader:
    """Excel file reader."""

    read_format = [".xlsx"]
    write_format = [".csv", ".tsv", ".xlsx"]

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Arguments
        ---------
        fp: str, pathlib.Path
            File path to the Excel file (.xlsx).

        Returns
        -------
        list:
            List with entries.
        """
        try:
            dfs = pd.read_excel(fp, sheet_name=None)
        except UnicodeDecodeError:
            dfs = pd.read_excel(fp, sheet_name=None, encoding="ISO-8859-1")

        best_sheet = None
        sheet_obj_val = -1
        wanted_columns = []
        for type_name, type_list in COLUMN_DEFINITIONS.items():
            wanted_columns.extend(type_list)

        for sheet_name in dfs:
            col_names = set([col.lower() for col in list(dfs[sheet_name])])
            obj_val = len(col_names & set(wanted_columns))
            if obj_val > sheet_obj_val:
                sheet_obj_val = obj_val
                best_sheet = sheet_name

        return _standardize_dataframe(dfs[best_sheet])



import asreview.webapp.tests.integration_tests.utils as utils
from asreview.webapp.authentication.models import User

ACCOUNT = {
    "email": "test4@user.org",
    "name": "Test User",
    "affiliation": "Utrecht University",
    "password": "@Secret1234!"
}


def test_change_profile(driver, url, database_uri):
    base_url = url
    driver.get(base_url)

    # SETUP  DATABASE
    session = utils.setup_database_session(database_uri)

    # clean database
    utils.clean_database(session)

    # assert we have no users
    assert len(session.query(User).all()) == 0

    # create account
    utils.create_account(driver, base_url, ACCOUNT)

    # assert we have a correct user
    assert len(session.query(User).all()) == 1
    user = session.query(User).first()
    session.close()
    assert user.email == ACCOUNT["email"]
    assert user.name == ACCOUNT["name"]
    assert user.affiliation == ACCOUNT["affiliation"]

    # sign in
    utils.sign_in(driver, base_url, ACCOUNT)

    # assert we're on the project dashboard
    assert utils.page_contains_text(driver, "Projects dashboard")

    # go to the profile page
    driver.get(base_url + "/profile")

    new_user_data = {
        "email": "casper@compunist.nl",
        "name": "Casper Kaandorp",
        "affiliation": "UU",
        "old_password": ACCOUNT["password"],
        "password": "uuuuUUUU111"
    }

    utils.fill_text_field_by_id(driver, "email", new_user_data["email"])
    utils.fill_text_field_by_id(driver, "name", new_user_data["name"])
    utils.fill_text_field_by_id(driver, "affiliation", new_user_data["affiliation"])
    utils.fill_text_field_by_id(driver, "oldPassword", new_user_data["old_password"])
    utils.fill_text_field_by_id(driver, "newPassword", new_user_data["password"])
    utils.fill_text_field_by_id(driver, "confirmPassword", new_user_data["password"])
    utils.click_element(driver, "button#save")

    # verify we're on the project dashboard again
    assert driver.current_url == base_url + "/projects"

    # assert we have an updated user
    assert len(session.query(User).all()) == 1
    user = session.query(User).first()
    assert user.email == new_user_data["email"]
    assert user.name == new_user_data["name"]
    assert user.affiliation == new_user_data["affiliation"]

    # log out
    utils.sign_out(driver)

    # log back in with new data
    utils.sign_in(driver, base_url, new_user_data)

    # check if we are on the project dashboard
    assert driver.current_url == base_url + "/projects"

    # close driver
    driver.close()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = []

import logging

import numpy as np

from asreview.models.balance.base import BaseBalance
from asreview.models.balance.double import DoubleBalance
from asreview.models.balance.double import _one_weight
from asreview.models.balance.double import _zero_weight
from asreview.models.balance.double import fill_training
from asreview.models.balance.double import random_round
from asreview.utils import get_random_state


class TripleBalance(BaseBalance):
    """Triple balance strategy (``triple``).

    Broken. Only for internal and experimental use.

    This divides the training data into three sets: included papers, excluded
    papers found with random sampling and papers found with max sampling. They
    are balanced according to formulas depending on the percentage of papers
    read in the dataset, the number of papers with random/max sampling etc.
    Works best for stochastic training algorithms. Reduces to both full
    sampling and undersampling with corresponding parameters.

    Arguments
    ---------
    a: float
        Governs the weight of the 1's. Higher values mean linearly more 1's
        in your training sample.
    alpha: float
        Governs the scaling the weight of the 1's, as a function of the
        ratio of ones to zeros. A positive value means that the lower the
        ratio of zeros to ones, the higher the weight of the ones.
    b: float
        Governs how strongly we want to sample depending on the total
        number of samples. A value of 1 means no dependence on the total
        number of samples, while lower values mean increasingly stronger
        dependence on the number of samples.
    beta: float
        Governs the scaling of the weight of the zeros depending on the
        number of samples. Higher values means that larger samples are more
        strongly penalizing zeros.
    c: float
        Value between one and zero that governs the weight of samples done
        with maximal sampling. Higher values mean higher weight.
    gamma: float
        Governs the scaling of the weight of the max samples as a function
        of the % of papers read. Higher values mean stronger scaling.
    """

    name = "triple"
    label = "Dynamic resampling (Triple)"

    def __init__(
        self,
        a=2.155,
        alpha=0.94,
        b=0.789,
        beta=1.0,
        c=0.835,
        gamma=2.0,
        shuffle=True,
        random_state=None,
    ):
        """Initialize the triple balance strategy."""

        super(TripleBalance, self).__init__()
        self.a = a
        self.alpha = alpha
        self.b = b
        self.beta = beta
        self.c = c
        self.gamma = gamma
        self.shuffle = shuffle
        self.fallback_model = DoubleBalance(
            a=a, alpha=alpha, b=b, beta=beta, random_state=random_state
        )
        self._random_state = get_random_state(random_state)

    def sample(self, X, y, train_idx, shared):
        """Resample the training data.

        Arguments
        ---------
        X: numpy.ndarray
            Complete feature matrix.
        y: numpy.ndarray
            Labels for all papers.
        train_idx: numpy.ndarray
            Training indices, that is all papers that have been reviewed.
        shared: dict
            Dictionary to share data between balancing models and other models.

        Returns
        -------
        numpy.ndarray,numpy.ndarray:
            X_train, y_train: the resampled matrix, labels.
        """
        max_idx = np.array(shared["query_src"].get("max", []), dtype=int)
        rand_idx = np.array([], dtype=int)
        for qtype in shared["query_src"]:
            if qtype != "max":
                rand_idx = np.append(rand_idx, shared["query_src"][qtype])

        rand_idx = rand_idx.astype(int)
        # Write them back for next round.
        if self.shuffle:
            self._random_state.shuffle(rand_idx)
            self._random_state.shuffle(max_idx)

        if len(rand_idx) == 0 or len(max_idx) == 0:
            logging.debug(
                "Warning: trying to use triple balance, but unable"
                f" to, because we have {len(max_idx)} max samples "
                f"and {len(rand_idx)} random samples."
            )
            return self.fallback_model.sample(X, y, train_idx, shared)

        # Split the idx into three groups: 1's, random 0's, max 0's.
        one_idx = train_idx[np.where(y[train_idx] == 1)]
        zero_max_idx = max_idx[np.where(y[max_idx] == 0)]
        zero_rand_idx = rand_idx[np.where(y[rand_idx] == 0)]

        if len(zero_rand_idx) == 0 or len(zero_max_idx) == 0:
            logging.debug(
                "Warning: trying to use triple balance, but unable "
                f"to, because we have {len(zero_max_idx)} zero max"
                f"samples and {len(zero_rand_idx)} random samples."
            )
            return self.fallback_model.sample(X, y, train_idx, shared)

        n_one = len(one_idx)
        n_zero_rand = len(zero_rand_idx)
        n_zero_max = len(zero_max_idx)
        n_samples = len(y)
        n_train = len(train_idx)

        # Get the distribution of 1's, and random 0's and max 0's.
        n_one_train, n_zero_rand_train, n_zero_max_train = _get_triple_dist(
            n_one,
            n_zero_rand,
            n_zero_max,
            n_samples,
            n_train,
            self.a,
            self.alpha,
            self.b,
            self.beta,
            self.c,
            self.gamma,
            self._random_state,
        )
        logging.debug(
            f"(1, 0_rand, 0_max) = ({n_one_train}, "
            f"{n_zero_rand_train}, {n_zero_max_train})"
        )

        one_train_idx = fill_training(one_idx, n_one_train, self._random_state)
        zero_rand_train_idx = fill_training(
            zero_rand_idx, n_zero_rand_train, self._random_state
        )
        zero_max_train_idx = fill_training(
            zero_max_idx, n_zero_max_train, self._random_state
        )

        all_idx = np.concatenate(
            [one_train_idx, zero_rand_train_idx, zero_max_train_idx]
        )
        self._random_state.shuffle(all_idx)

        return X[all_idx], y[all_idx]

    def full_hyper_space(self):
        from hyperopt import hp

        parameter_space = {
            "bal_a": hp.lognormal("bal_a", 0, 1),
            "bal_alpha": hp.uniform("bal_alpha", 0, 2),
            "bal_b": hp.uniform("bal_b", 0, 1),
            # "bal_zero_beta": hp.uniform("bal_zero_beta", 0, 2),
            "bal_c": hp.uniform("bal_c", 0, 1),
            # "bal_zero_max_gamma": hp.uniform("bal_zero_max_gamma", 0.01, 2)
        }
        return parameter_space, {}


def _zero_max_weight(fraction_read, c, gamma):
    """
    Get the weight ratio between ones and zeros.

    Parameters
    ----------
    beta:
        Exponent governing decay of 1's to 0's.
    delta:
        Asymptotic ratio for infinite number of samples.

    Returns
    -------
    float:
        Weight ratio between 1's and 0's
    """
    weight = 1 - (1 - c) * (1 - fraction_read) ** gamma
    return weight


def _get_triple_dist(
    n_one,
    n_zero_rand,
    n_zero_max,
    n_samples,
    n_train,
    one_a,
    one_alpha,
    zero_b,
    zero_beta,
    zero_max_c,
    zero_max_gamma,
    random_state,
):
    "Get the number of 1's, random 0's and max 0's in each mini epoch."
    n_zero = n_zero_rand + n_zero_max
    n_read = n_one + n_zero
    one_weight = _one_weight(n_one, n_zero, one_a, one_alpha)
    zero_weight = _zero_weight(n_read, zero_b, zero_beta)
    zero_max_weight = _zero_max_weight(n_read / n_samples, zero_max_c, zero_max_gamma)

    tot_zo_weight = one_weight * n_one + zero_weight * n_zero

    n_one_train = random_round(
        one_weight * n_one * n_train / tot_zo_weight, random_state
    )
    n_one_train = max(1, min(n_train - 2, n_one_train))
    n_zero_train = n_train - n_one_train

    tot_rm_weight = 1 * n_zero_rand + zero_max_weight * n_zero_max
    n_zero_rand_train = random_round(
        n_zero_train * 1 * n_zero_rand / tot_rm_weight, random_state
    )
    n_zero_rand_train = max(1, min(n_zero_rand - 1, n_zero_rand_train))
    n_zero_max_train = n_zero_train - n_zero_rand_train

    return n_one_train, n_zero_rand_train, n_zero_max_train



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path
from urllib.request import urlretrieve
from uuid import uuid4

import datahugger
import numpy as np
import pandas as pd
from flask import Blueprint
from flask import abort
from flask import current_app
from flask import jsonify
from flask import request
from flask import send_file
from flask_login import current_user
from flask_login import login_required
from sqlalchemy import and_
from werkzeug.exceptions import InternalServerError
from werkzeug.utils import secure_filename

from asreview.config import DEFAULT_BALANCE_STRATEGY
from asreview.config import DEFAULT_FEATURE_EXTRACTION
from asreview.config import DEFAULT_MODEL
from asreview.config import DEFAULT_QUERY_STRATEGY
from asreview.config import LABEL_NA
from asreview.config import PROJECT_MODE_EXPLORE
from asreview.config import PROJECT_MODE_SIMULATE
from asreview.data.statistics import n_duplicates
from asreview.datasets import DatasetManager
from asreview.exceptions import BadFileFormatError
from asreview.io import list_readers
from asreview.io import list_writers
from asreview.models.balance import get_balance_model
from asreview.models.balance import list_balance_strategies
from asreview.models.classifiers import get_classifier
from asreview.models.classifiers import list_classifiers
from asreview.models.feature_extraction import get_feature_model
from asreview.models.feature_extraction import list_feature_extraction
from asreview.models.query import get_query_model
from asreview.models.query import list_query_strategies
from asreview.project import ASReviewProject
from asreview.project import ProjectNotFoundError
from asreview.project import get_project_path
from asreview.project import is_v0_project
from asreview.project import open_state
from asreview.search import SearchError
from asreview.search import fuzzy_find
from asreview.settings import ASReviewSettings
from asreview.state.errors import StateError
from asreview.state.errors import StateNotFoundError
from asreview.state.sql_converter import upgrade_asreview_project_file
from asreview.state.sql_converter import upgrade_project_config
from asreview.utils import _entry_points
from asreview.utils import _get_executable
from asreview.utils import _get_filename_from_url
from asreview.utils import asreview_path
from asreview.utils import list_reader_names
from asreview.webapp import DB
from asreview.webapp.authentication.decorators import current_user_projects
from asreview.webapp.authentication.decorators import project_authorization
from asreview.webapp.authentication.models import Project

bp = Blueprint("api", __name__, url_prefix="/api")


# error handlers
@bp.errorhandler(ValueError)
def value_error(e):
    message = str(e) if str(e) else "Incorrect value."
    logging.error(message)
    return jsonify(message=message), 400


@bp.errorhandler(ProjectNotFoundError)
def project_not_found(e):
    message = str(e) if str(e) else "Project not found."
    logging.error(message)
    return jsonify(message=message), 404


@bp.errorhandler(InternalServerError)
def error_500(e):
    original = getattr(e, "original_exception", None)

    if original is None or str(e.original_exception) == "":
        # direct 500 error, such as abort(500)
        logging.error(e)
        return jsonify(message="Whoops, something went wrong."), 500

    # wrapped unhandled error
    logging.error(e.original_exception)
    return jsonify(message=str(e.original_exception)), 500


# routes
@bp.route("/projects", methods=["GET"])
@login_required
@current_user_projects
def api_get_projects(projects):  # noqa: F401
    """"""
    project_info = []

    # for project, owner_id in zip(projects, owner_ids):
    for project in projects:
        try:
            project_config = project.config

            if not current_app.config.get("LOGIN_DISABLED", False):
                project_config["owner_id"] = current_user.id

            # upgrade info of v0 projects
            if project_config["version"].startswith("0"):
                project_config = upgrade_project_config(project_config)
                project_config["projectNeedsUpgrade"] = True

            logging.info("Project found: {}".format(project_config["id"]))
            project_info.append(project_config)

        except Exception as err:
            logging.error(err)

    # sort the projects based on created_at_unix
    project_info = sorted(
        project_info,
        key=lambda y: (y["created_at_unix"] is not None, y["created_at_unix"]),
        reverse=True,
    )

    response = jsonify({"result": project_info})

    return response


@bp.route("/projects/stats", methods=["GET"])
@login_required
@current_user_projects
def api_get_projects_stats(projects):  # noqa: F401
    """Get dashboard statistics of all projects"""

    stats_counter = {"n_in_review": 0, "n_finished": 0, "n_setup": 0}

    for project in projects:
        project_config = project.config

        # upgrade info of v0 projects
        if project_config["version"].startswith("0"):
            project_config = upgrade_project_config(project_config)
            project_config["projectNeedsUpgrade"] = True

        # get dashboard statistics
        try:
            if project_config["reviews"][0]["status"] == "review":
                stats_counter["n_in_review"] += 1
            elif project_config["reviews"][0]["status"] == "finished":
                stats_counter["n_finished"] += 1
            else:
                stats_counter["n_setup"] += 1
        except Exception:
            stats_counter["n_setup"] += 1

    return jsonify({"result": stats_counter})


@bp.route("/projects/info", methods=["POST"])
@login_required
def api_init_project():  # noqa: F401
    """Initialize a new project"""

    project_mode = request.form["mode"]
    project_title = request.form["name"]
    project_description = request.form["description"]
    project_authors = request.form["authors"]

    # get a unique project id
    project_id = uuid4().hex

    # get path of this project
    project_path = get_project_path(project_id)

    project = ASReviewProject.create(
        project_path,
        project_id=project_id,
        project_mode=project_mode,
        project_name=project_title,
        project_description=project_description,
        project_authors=project_authors,
    )

    if current_app.config.get("LOGIN_DISABLED", False):
        return jsonify(project.config), 201

    # create a database entry for this project
    current_user.projects.append(Project(project_id=project_id))
    DB.session.commit()

    return jsonify(project.config), 201


@bp.route("/dataset_readers", methods=["GET"])
@login_required
def api_list_data_readers():
    """Get the list of available data readers and read formats."""
    payload = {"result": []}
    for e in _entry_points(group="asreview.readers"):
        payload["result"].append(
            {
                "extension": e.name
            }
        )
    return jsonify(payload)


@bp.route("/projects/<project_id>/upgrade_if_old", methods=["GET"])
@login_required
@project_authorization
def api_upgrade_project_if_old(project):
    """Get upgrade project if it is v0.x"""

    if not project.config["version"].startswith("0"):
        response = jsonify(message="Can only convert v0.x projects.")
        return response, 400

    # errors are handled by the InternalServerError
    upgrade_asreview_project_file(project.project_path)

    response = jsonify({"success": True})
    return response


@bp.route("/projects/<project_id>/info", methods=["GET"])
@login_required
@project_authorization
def api_get_project_info(project):  # noqa: F401
    """"""
    project_config = project.config

    # upgrade info of v0 projects
    if project_config["version"].startswith("0"):
        project_config = upgrade_project_config(project_config)
        project_config["projectNeedsUpgrade"] = True

    if current_app.config.get("LOGIN_DISABLED", False):
        return jsonify(project_config)

    db_project = Project.query.filter(
        Project.project_id == project.config.get("id", 0)
    ).one_or_none()
    if db_project:
        project_config["ownerId"] = db_project.owner_id

    return jsonify(project_config)


@bp.route("/projects/<project_id>/info", methods=["PUT"])
@login_required
@project_authorization
def api_update_project_info(project):  # noqa: F401
    """Update project info"""

    project.update_config(
        mode=request.form["mode"],
        name=request.form["name"],
        description=request.form["description"],
        authors=request.form["authors"],
    )

    return api_get_project_info(project.project_id)


@bp.route("/datasets", methods=["GET"])
@login_required
def api_demo_data_project():  # noqa: F401
    """"""

    subset = request.args.get("subset", None)

    manager = DatasetManager()

    if subset == "plugin":
        try:
            result_datasets = manager.list(
                exclude=[
                    "builtin", "synergy", "benchmark", "benchmark-nature"
                ]
            )

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to load plugin datasets."), 500

    elif subset == "benchmark":
        try:
            # collect the datasets metadata
            result_datasets = manager.list(include=["synergy", "benchmark-nature"])

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to load benchmark datasets."), 500

    else:
        response = jsonify(message="demo-data-loading-failed")

        return response, 400

    payload = {"result": result_datasets}
    response = jsonify(payload)
    return response


@bp.route("/projects/<project_id>/data", methods=["POST", "PUT"])
@login_required
@project_authorization
def api_upload_data_to_project(project):  # noqa: F401
    """"""

    # get the project config to modify behavior of dataset
    project_config = project.config

    # remove old dataset if present
    if "dataset_path" in project_config and project_config["dataset_path"] is not None:
        logging.warning("Removing old dataset and adding new dataset.")
        project.remove_dataset()

    # create dataset folder if not present
    Path(project.project_path, "data").mkdir(exist_ok=True)

    if request.form.get("plugin", None):
        ds = DatasetManager().find(request.form["plugin"])
        filename = ds.filename
        ds.to_file(Path(project.project_path, "data", filename))

    elif request.form.get("benchmark", None):
        ds = DatasetManager().find(request.form["benchmark"])
        filename = ds.filename
        ds.to_file(Path(project.project_path, "data", filename))

    elif request.form.get("url", None):
        url = request.form.get("url")

        # check if url value is actually DOI without netloc
        if url.startswith("10."):
            url = f"https://doi.org/{url}"

        filename = _get_filename_from_url(url)

        if bool(request.form.get("validate", None)):
            reader_keys = list_reader_names()

            if (
                filename
                and Path(filename).suffix
                and Path(filename).suffix in reader_keys
            ):
                return jsonify(files=[{"link": url, "name": filename}]), 201
            elif filename and not Path(filename).suffix:
                raise BadFileFormatError("Can't determine file format.")
            else:
                try:
                    # get file list from datahugger
                    dh = datahugger.info(url)
                    files = dh.files.copy()

                    for i, f in enumerate(files):
                        files[i]["disabled"] = (
                            Path(files[i]["name"]).suffix not in reader_keys
                        )

                    return jsonify(files=files), 201
                except Exception:
                    raise BadFileFormatError("Can't retrieve files.")

        try:
            urlretrieve(url, Path(project.project_path, "data") / filename)
        except Exception as err:
            logging.error(err)
            message = f"Can't retrieve data from URL {url}."

            return jsonify(message=message), 400

    elif "file" in request.files:
        data_file = request.files["file"]

        # check the file is file is in a correct format
        # check_dataset(data_file)
        try:
            filename = secure_filename(data_file.filename)
            fp_data = Path(project.project_path, "data") / filename

            # save the file
            data_file.save(str(fp_data))

        except Exception as err:
            logging.error(err)

            response = jsonify(message=f"Failed to import file '{filename}'. {err}")

            return response, 400
    else:
        response = jsonify(message="No file or dataset found to import.")
        return response, 400

    data_path = Path(project.project_path, "data") / filename

    try:
        # add the file to the project
        project.add_dataset(data_path.name)

    # Bad format. TODO{Jonathan} Return informative message with link.
    except Exception as err:

        try:
            project.remove_dataset()
        except Exception:
            pass

        message = f"Failed to import file '{filename}'. {err}"
        return jsonify(message=message), 400

    response = jsonify({"project_id": project.project_id})

    return response


@bp.route("/projects/<project_id>/data", methods=["GET"])
@login_required
@project_authorization
def api_get_project_data(project):  # noqa: F401
    """"""

    try:
        # get statistics of the dataset
        as_data = project.read_data()

        statistics = {
            "n_rows": as_data.df.shape[0],
            "n_cols": as_data.df.shape[1],
            "n_duplicates": n_duplicates(as_data),
            "filename": Path(project.config["dataset_path"]).stem,
        }

    except FileNotFoundError as err:
        logging.info(err)
        statistics = {"filename": None}

    return jsonify(statistics)


@bp.route("/projects/<project_id>/dataset_writer", methods=["GET"])
@login_required
@project_authorization
def api_list_dataset_writers(project):
    """List the name and label of available dataset writer"""

    fp_data = Path(project.config["dataset_path"])

    readers = list_readers()
    writers = list_writers()

    # get write format for the data file
    write_format = None
    for c in readers:
        if fp_data.suffix in c.read_format:
            if write_format is None:
                write_format = c.write_format

    # get available writers
    payload = {"result": []}
    for c in writers:
        payload["result"].append(
            {
                "enabled": True if c.write_format in write_format else False,
                "name": c.name,
                "label": c.label,
                "caution": c.caution if hasattr(c, "caution") else None,
            }
        )

    if not payload["result"]:
        return (
            jsonify(message=f"No dataset writer available for {fp_data.suffix} file."),
            500,
        )

    # remove duplicate writers
    payload["result"] = [
        i
        for n, i in enumerate(payload["result"])
        if i not in payload["result"][(n + 1) :]
    ]

    return jsonify(payload)


@bp.route("/projects/<project_id>/search", methods=["GET"])
@login_required
@project_authorization
def api_search_data(project):  # noqa: F401
    """Search for papers"""
    q = request.args.get("q", default=None, type=str)
    max_results = request.args.get("n_max", default=10, type=int)

    payload = {"result": []}
    if q:
        # read the dataset
        as_data = project.read_data()

        # read record_ids of labels from state
        with open_state(project.project_path) as s:
            labeled_record_ids = s.get_dataset(["record_id"])["record_id"].to_list()

        try:
            # search for the keywords
            result_idx = fuzzy_find(
                as_data,
                q,
                max_return=max_results,
                exclude=labeled_record_ids,
                by_index=True,
            )
        except SearchError as err:
            raise ValueError(err) from err

        for record in as_data.record(result_idx):

            payload["result"].append(
                {
                    "id": int(record.record_id),
                    "title": record.title,
                    "abstract": record.abstract,
                    "authors": record.authors,
                    "keywords": record.keywords,
                    "included": -1,
                    "label_from_dataset": int(record.included),
                }
            )

    return jsonify(payload)


@bp.route("/projects/<project_id>/labeled", methods=["GET"])
@login_required
@project_authorization
def api_get_labeled(project):  # noqa: F401
    """Get all papers classified as labeled documents"""

    page = request.args.get("page", default=None, type=int)
    per_page = request.args.get("per_page", default=20, type=int)
    subset = request.args.getlist("subset")
    latest_first = request.args.get("latest_first", default=1, type=int)

    with open_state(project.project_path) as s:
        data = s.get_dataset(["record_id", "label", "query_strategy", "notes"])
        data["prior"] = (data["query_strategy"] == "prior").astype(int)

    if any(s in subset for s in ["relevant", "included"]):
        data = data[data["label"] == 1]
    elif any(s in subset for s in ["irrelevant", "excluded"]):
        data = data[data["label"] == 0]
    else:
        data = data[~data["label"].isnull()]

    if "note" in subset:
        data = data[~data["notes"].isnull()]

    if "prior" in subset:
        data = data[data["prior"] == 1]

    if latest_first == 1:
        data = data.iloc[::-1]

    # count labeled records and max pages
    count = len(data)
    if count == 0:
        payload = {
            "count": 0,
            "next_page": None,
            "previous_page": None,
            "result": [],
        }
        response = jsonify(payload)

        return response

    max_page_calc = divmod(count, per_page)
    if max_page_calc[1] == 0:
        max_page = max_page_calc[0]
    else:
        max_page = max_page_calc[0] + 1

    if page is not None:
        # slice out records on specific page
        if page <= max_page:
            idx_start = page * per_page - per_page
            idx_end = page * per_page
            data = data.iloc[idx_start:idx_end, :].copy()
        else:
            return abort(404)

        # set next & previous page
        if page < max_page:
            next_page = page + 1
            if page > 1:
                previous_page = page - 1
            else:
                previous_page = None
        else:
            next_page = None
            previous_page = page - 1
    else:
        next_page = None
        previous_page = None

    records = project.read_data().record(data["record_id"])

    payload = {
        "count": count,
        "next_page": next_page,
        "previous_page": previous_page,
        "result": [],
    }
    for i, record in zip(data.index.tolist(), records):
        payload["result"].append(
            {
                "id": int(record.record_id),
                "title": record.title,
                "abstract": record.abstract,
                "authors": record.authors,
                "keywords": record.keywords,
                "doi": record.doi,
                "url": record.url,
                "included": int(data.loc[i, "label"]),
                "note": data.loc[i, "notes"],
                "prior": int(data.loc[i, "prior"]),
            }
        )

    return jsonify(payload)


@bp.route("/projects/<project_id>/labeled_stats", methods=["GET"])
@login_required
@project_authorization
def api_get_labeled_stats(project):  # noqa: F401
    """Get all papers classified as prior documents"""

    try:
        with open_state(project.project_path) as s:
            data = s.get_dataset(["label", "query_strategy"])
            data_prior = data[data["query_strategy"] == "prior"]

        return jsonify(
            {
                "n": len(data),
                "n_inclusions": sum(data["label"] == 1),
                "n_exclusions": sum(data["label"] == 0),
                "n_prior": len(data_prior),
                "n_prior_inclusions": sum(data_prior["label"] == 1),
                "n_prior_exclusions": sum(data_prior["label"] == 0),
            }
        )
    except StateNotFoundError:
        return jsonify(
            {
                "n": 0,
                "n_inclusions": 0,
                "n_exclusions": 0,
                "n_prior": 0,
                "n_prior_inclusions": 0,
                "n_prior_exclusions": 0,
            }
        )


@bp.route("/projects/<project_id>/prior_random", methods=["GET"])
@login_required
@project_authorization
def api_random_prior_papers(project):  # noqa: F401
    """Get a selection of random records.

    This set of records is extracted from the pool, but without
    the already labeled items.
    """

    # get the number of records to return
    n = request.args.get("n", default=5, type=int)
    # get the subset of records to return (for exploration and simulation mode)
    subset = request.args.get("subset", default=None, type=str)

    with open_state(project.project_path) as state:
        pool = state.get_pool().values

    as_data = project.read_data()

    payload = {"result": []}

    if subset in ["relevant", "included"]:

        if as_data.labels is None:
            return jsonify(payload)

        rel_indices = as_data.df[as_data.labels == 1].index.values
        rel_indices_pool = np.intersect1d(pool, rel_indices)

        if len(rel_indices_pool) == 0:
            return jsonify(payload)
        elif n > len(rel_indices_pool):
            rand_pool_relevant = np.random.choice(
                rel_indices_pool, len(rel_indices_pool), replace=False
            )
        else:
            rand_pool = np.random.choice(pool, n, replace=False)
            rand_pool_relevant = np.random.choice(rel_indices_pool, n, replace=False)

        try:
            relevant_records = as_data.record(rand_pool_relevant)
        except Exception as err:
            logging.error(err)
            return jsonify(message=f"Failed to load random records. {err}"), 500

        for rr in relevant_records:
            payload["result"].append(
                {
                    "id": int(rr.record_id),
                    "title": rr.title,
                    "abstract": rr.abstract,
                    "authors": rr.authors,
                    "keywords": rr.keywords,
                    "included": None,
                    "label_from_dataset": 1,
                }
            )

    elif subset in ["irrelevant", "excluded"]:

        if as_data.labels is None:
            return jsonify(payload)

        irrel_indices = as_data.df[as_data.labels == 0].index.values
        irrel_indices_pool = np.intersect1d(pool, irrel_indices)

        if len(irrel_indices_pool) == 0:
            return jsonify(payload)
        elif n > len(irrel_indices_pool):
            rand_pool_irrelevant = np.random.choice(
                irrel_indices_pool, len(irrel_indices_pool), replace=False
            )
        else:
            rand_pool_irrelevant = np.random.choice(
                irrel_indices_pool, n, replace=False
            )

        try:
            irrelevant_records = as_data.record(rand_pool_irrelevant)
        except Exception as err:
            logging.error(err)
            return jsonify(message=f"Failed to load random records. {err}"), 500

        for ir in irrelevant_records:
            payload["result"].append(
                {
                    "id": int(ir.record_id),
                    "title": ir.title,
                    "abstract": ir.abstract,
                    "authors": ir.authors,
                    "keywords": ir.keywords,
                    "included": None,
                    "label_from_dataset": 0,
                }
            )

    elif subset == "not_seen":
        # Fetch records that are not seen
        unlabeled_indices = as_data.df[as_data.labels == LABEL_NA] \
            .index.values
        unlabeled_indices_pool = np.intersect1d(pool, unlabeled_indices)

        if len(unlabeled_indices_pool) == 0:
            return jsonify(payload)
        elif n > len(unlabeled_indices_pool):
            rand_pool_unlabeled = np.random.choice(
                unlabeled_indices_pool,
                len(unlabeled_indices_pool),
                replace=False
            )
        else:
            rand_pool_unlabeled = np.random.choice(
                unlabeled_indices_pool,
                n,
                replace=False
            )

        try:
            unlabeled_records = as_data.record(rand_pool_unlabeled)
        except Exception as err:
            logging.error(err)
            return jsonify(message=f"Failed to load 'not seen' records. {err}"), 500

        for record in unlabeled_records:
            payload["result"].append(
                {
                    "id": int(record.record_id),
                    "title": record.title,
                    "abstract": record.abstract,
                    "authors": record.authors,
                    "keywords": record.keywords,
                    "included": None,
                    "label_from_dataset": -1,
                }
            )
    else:
        if len(pool) == 0:
            return jsonify(payload)
        elif n > len(pool):
            rand_pool = np.random.choice(pool, len(pool), replace=False)
        else:
            rand_pool = np.random.choice(pool, n, replace=False)

        try:
            records = as_data.record(rand_pool)
        except Exception as err:
            logging.error(err)
            return jsonify(message=f"Failed to load random records. {err}"), 500

        for r in records:
            payload["result"].append(
                {
                    "id": int(r.record_id),
                    "title": r.title,
                    "abstract": r.abstract,
                    "authors": r.authors,
                    "keywords": r.keywords,
                    "included": None,
                    "label_from_dataset": None,
                }
            )

    return jsonify(payload)


@bp.route("/algorithms", methods=["GET"])
@login_required
def api_list_algorithms():
    """List the names and labels of available algorithms"""

    classes = [
        list_balance_strategies(),
        list_classifiers(),
        list_feature_extraction(),
        list_query_strategies(),
    ]

    payload = {
        "balance_strategy": [],
        "classifier": [],
        "feature_extraction": [],
        "query_strategy": [],
    }

    for c, key in zip(classes, payload.keys()):
        for method in c:
            if hasattr(method, "label"):
                payload[key].append({"name": method.name, "label": method.label})
            else:
                payload[key].append({"name": method.name, "label": method.name})

    return jsonify(payload)


@bp.route("/projects/<project_id>/algorithms", methods=["GET"])
@login_required
@project_authorization
def api_get_algorithms(project):  # noqa: F401
    default_payload = {
        "model": DEFAULT_MODEL,
        "feature_extraction": DEFAULT_FEATURE_EXTRACTION,
        "query_strategy": DEFAULT_QUERY_STRATEGY,
        "balance_strategy": DEFAULT_BALANCE_STRATEGY,
    }

    # check if there were algorithms stored in the state file
    try:
        with open_state(project.project_path) as state:
            if state.settings is not None:
                payload = {
                    "model": state.settings.model,
                    "feature_extraction": state.settings.feature_extraction,
                    "query_strategy": state.settings.query_strategy,
                    "balance_strategy": state.settings.balance_strategy,
                }
            else:
                payload = default_payload
    except StateNotFoundError:
        payload = default_payload

    return jsonify(payload)


@bp.route("/projects/<project_id>/algorithms", methods=["POST"])
@login_required
@project_authorization
def api_set_algorithms(project):  # noqa: F401
    # TODO@{Jonathan} validate model choice on server side
    ml_model = request.form.get("model", None)
    ml_query_strategy = request.form.get("query_strategy", None)
    ml_balance_strategy = request.form.get("balance_strategy", None)
    ml_feature_extraction = request.form.get("feature_extraction", None)

    # create a new settings object from arguments
    # only used if state file is not present
    asreview_settings = ASReviewSettings(
        model=ml_model,
        query_strategy=ml_query_strategy,
        balance_strategy=ml_balance_strategy,
        feature_extraction=ml_feature_extraction,
        model_param=get_classifier(ml_model).param,
        query_param=get_query_model(ml_query_strategy).param,
        balance_param=get_balance_model(ml_balance_strategy).param,
        feature_param=get_feature_model(ml_feature_extraction).param,
    )

    # save the new settings to the state file
    with open_state(project.project_path, read_only=False) as state:
        state.settings = asreview_settings

    response = jsonify({"success": True})

    return response


@bp.route("/projects/<project_id>/start", methods=["POST"])
@login_required
@project_authorization
def api_start(project):  # noqa: F401
    """Start training of first model or simulation."""

    # the project is a simulation project
    if project.config["mode"] == PROJECT_MODE_SIMULATE:
        # get priors
        with open_state(project.project_path) as s:
            priors = s.get_priors()["record_id"].tolist()

        logging.info("Start simulation")

        try:
            datafile = project.config["dataset_path"]
            logging.info("Project data file found: {}".format(datafile))

            # start simulation
            py_exe = _get_executable()
            run_command = (
                [
                    # get executable
                    py_exe,
                    # get module
                    "-m",
                    "asreview",
                    # run simulation via cli
                    "simulate",
                    # specify dataset
                    "",
                    # specify prior indices
                    "--prior_idx",
                ]
                + list(map(str, priors))
                + [
                    # specify state file
                    "--state_file",
                    str(project.project_path),
                    # specify write interval
                    "--write_interval",
                    "100",
                ]
            )
            subprocess.Popen(run_command)

        except Exception as err:
            logging.error(err)
            message = f"Failed to get data file. {err}"
            return jsonify(message=message), 400

    # the project is an oracle or explore project
    else:
        logging.info("Train first iteration of model")
        try:
            # start training the model
            py_exe = _get_executable()
            run_command = [
                # get executable
                py_exe,
                # get module
                "-m",
                "asreview",
                # train the model via cli
                "web_run_model",
                # specify project id
                str(project.project_path),
                # output the error of the first model
                "--output_error",
            ]
            subprocess.Popen(run_command)

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to train the model."), 500

    response = jsonify({"success": True})

    return response


@bp.route("/projects/<project_id>/status", methods=["GET"])
@login_required
@project_authorization
def api_get_status(project):  # noqa: F401
    """Check the status of the review"""

    try:
        status = project.reviews[0]["status"]
    except Exception:
        status = None

    if status == "error":
        error_path = project.project_path / "error.json"
        if error_path.exists():
            logging.error("Error on training")
            with open(error_path, "r") as f:
                error_message = json.load(f)["message"]

            raise Exception(error_message)

    response = jsonify({"status": status})

    return response


@bp.route("/projects/<project_id>/status", methods=["PUT"])
@login_required
@project_authorization
def api_status_update(project):
    """Update the status of the review.

    The following status updates are allowed for
    oracle and explore:
    - `review` to `finished`
    - `finished` to `review` if not pool empty
    - `error` to `setup`

    The following status updates are allowed for simulate
    - `error` to `setup`

    Status updates by the user are not allowed in simulation
    mode.

    """

    status = request.form.get("status", type=str)

    current_status = project.config["reviews"][0]["status"]
    mode = project.config["mode"]

    if current_status == "error" and status == "setup":
        project.remove_error(status=status)

        response = jsonify({"success": True})

        return response

    if mode == PROJECT_MODE_SIMULATE:
        raise ValueError("Not possible to update status of simulation project.")
    else:
        if current_status == "review" and status == "finished":
            project.update_review(status=status)
        elif current_status == "finished" and status == "review":
            project.update_review(status=status)
            # ideally, also check here for empty pool
        else:
            raise ValueError(
                f"Not possible to update status from {current_status} to {status}"
            )

        response = jsonify({"success": True})

        return response


@bp.route("/projects/import_project", methods=["POST"])
@login_required
def api_import_project():
    """Import project"""

    # raise error if file not given
    if "file" not in request.files:
        response = jsonify(message="No ASReview file found to import.")
        return response, 400

    try:
        project = ASReviewProject.load(
            request.files["file"],
            asreview_path(),
            safe_import=True
        )

    except Exception as err:
        logging.error(err)
        raise ValueError("Failed to import project.")

    if current_app.config.get("LOGIN_DISABLED", False):
        return jsonify(project.config)

    # create a database entry for this project
    current_user.projects.append(
        Project(project_id=project.config.get("id"))
    )
    project.config["owner_id"] = current_user.id
    DB.session.commit()

    return jsonify(project.config)


@bp.route("/projects/<project_id>/export_dataset", methods=["GET"])
@login_required
@project_authorization
def api_export_dataset(project):
    """Export dataset with relevant/irrelevant labels"""

    # get the export args
    file_format = request.args.get("file_format", None)
    dataset_label = request.args.get("dataset_label", default="all")

    # create temporary folder to store exported dataset
    tmp_path = tempfile.TemporaryDirectory()
    tmp_path_dataset = Path(tmp_path.name, f"export_dataset.{file_format}")

    try:
        # get labels and ranking from state file
        with open_state(project.project_path) as s:
            pool, labeled, pending = s.get_pool_labeled_pending()
            # get state dataset for accessing notes
            state_df = s.get_dataset().set_index("record_id")

        included = labeled[labeled["label"] == 1]
        excluded = labeled[labeled["label"] != 1]

        if dataset_label == "relevant":
            export_order = included["record_id"].to_list()
            labeled = included
        else:
            export_order = (
                included["record_id"].to_list()
                + pending.to_list()
                + pool.to_list()
                + excluded["record_id"].to_list()
            )

        # get writer corresponding to specified file format
        writers = list_writers()
        writer = None
        for c in writers:
            if writer is None:
                if c.name == file_format:
                    writer = c

        # read the dataset into a ASReview data object
        as_data = project.read_data()

        # Add a new column 'is_prior' to the dataset
        if "asreview_prior" in as_data.df:
            as_data.df.drop("asreview_prior", axis=1, inplace=True)

        state_df["asreview_prior"] = state_df.query_strategy.eq("prior").astype("Int64")
        as_data.df = as_data.df.join(state_df["asreview_prior"], on="record_id")

        # Adding Notes from State file to the exported dataset
        # Check if exported_notes column already exists due to multiple screenings
        screening = 0
        for col in as_data.df:
            if col == "exported_notes":
                screening = 0
            elif col.startswith("exported_notes"):
                try:
                    screening = int(col.split("_")[2])
                except IndexError:
                    screening = 0
        screening += 1

        state_df.rename(
            columns={
                "notes": f"exported_notes_{screening}",
            },
            inplace=True,
        )

        as_data.df = as_data.df.join(
            state_df[f"exported_notes_{screening}"], on="record_id"
        )

        # keep labels in exploration mode
        keep_old_labels = project.config["mode"] == PROJECT_MODE_EXPLORE

        as_data.to_file(
            fp=tmp_path_dataset,
            labels=labeled.values.tolist(),
            ranking=export_order,
            writer=writer,
            keep_old_labels=keep_old_labels,
        )

        return send_file(
            tmp_path_dataset,
            as_attachment=True,
            max_age=0,
        )

    except Exception as err:
        raise Exception(f"Failed to export the {file_format} dataset. {err}")


@bp.route("/projects/<project_id>/export_project", methods=["GET"])
@login_required
@project_authorization
def export_project(project):
    """Export the project file.

    The ASReview project file is a file with .asreview extension.
    The ASReview project file is a zipped file and contains
    all information to continue working on the project as well
    as the original dataset.
    """

    # create a temp folder to zip
    tmpdir = tempfile.TemporaryDirectory()
    tmpfile = Path(tmpdir.name, project.project_id).with_suffix(".asreview")

    logging.info("Saving project (temporary) to %s", tmpfile)
    project.export(tmpfile)

    return send_file(
        tmpfile,
        as_attachment=True,
        max_age=0,
    )


def _get_stats(project, include_priors=False):
    if is_v0_project(project.project_path):
        json_fp = Path(project.project_path, "result.json")

        # Check if the v0 project is in review.
        if json_fp.exists():
            with open(json_fp, "r") as f:
                s = json.load(f)

            # Get the labels.
            labels = np.array(
                [
                    int(sample_data[1])
                    for query in range(len(s["results"]))
                    for sample_data in s["results"][query]["labelled"]
                ]
            )

            # Get the record table.
            data_hash = list(s["data_properties"].keys())[0]
            record_table = s["data_properties"][data_hash]["record_table"]

            n_records = len(record_table)

        # No result found.
        else:
            labels = np.array([])
            n_records = 0
    else:
        # Check if there is a review started in the project.
        try:
            # get label history
            with open_state(project.project_path) as s:
                if (
                    project.config["reviews"][0]["status"] == "finished"
                    and project.config["mode"] == PROJECT_MODE_SIMULATE
                ):
                    labels = _get_labels(s, priors=include_priors)
                else:
                    labels = s.get_labels(priors=include_priors)

                n_records = len(s.get_record_table())

        # No state file found or not init.
        except (StateNotFoundError, StateError):
            labels = np.array([])
            n_records = 0

    n_included = int(sum(labels == 1))
    n_excluded = int(sum(labels == 0))

    if n_included > 0:
        n_since_last_relevant = int(labels.tolist()[::-1].index(1))
    else:
        n_since_last_relevant = 0

    return {
        "n_included": n_included,
        "n_excluded": n_excluded,
        "n_since_last_inclusion": n_since_last_relevant,
        "n_papers": n_records,
        "n_pool": n_records - n_excluded - n_included,
    }


def _get_labels(state_obj, priors=False):
    # get the number of records
    n_records = state_obj.n_records

    # get the labels
    labels = state_obj.get_labels(priors=priors).to_list()

    # if less labels than records, fill with 0
    if len(labels) < n_records:
        labels += [0] * (n_records - len(labels))
        labels = pd.Series(labels)

    return labels


@bp.route("/projects/<project_id>/progress", methods=["GET"])
@login_required
@project_authorization
def api_get_progress_info(project):  # noqa: F401
    """Get progress statistics of a project"""

    include_priors = request.args.get("priors", True, type=bool)

    response = jsonify(_get_stats(project, include_priors=include_priors))

    # return a success response to the client.
    return response


@bp.route("/projects/<project_id>/progress_density", methods=["GET"])
@login_required
@project_authorization
def api_get_progress_density(project):
    """Get progress density of a project"""

    include_priors = request.args.get("priors", False, type=bool)

    # get label history
    with open_state(project.project_path) as s:
        if (
            project.config["reviews"][0]["status"] == "finished"
            and project.config["mode"] == PROJECT_MODE_SIMULATE
        ):
            data = _get_labels(s, priors=include_priors)
        else:
            data = s.get_labels(priors=include_priors)

    # create a dataset with the rolling mean of every 10 papers
    df = (
        data.to_frame(name="Relevant")
        .reset_index(drop=True)
        .rolling(10, min_periods=1)
        .mean()
    )
    df["Total"] = df.index + 1

    # transform mean(percentage) to number
    for i in range(0, len(df)):
        if df.loc[i, "Total"] < 10:
            df.loc[i, "Irrelevant"] = (1 - df.loc[i, "Relevant"]) * df.loc[i, "Total"]
            df.loc[i, "Relevant"] = df.loc[i, "Total"] - df.loc[i, "Irrelevant"]
        else:
            df.loc[i, "Irrelevant"] = (1 - df.loc[i, "Relevant"]) * 10
            df.loc[i, "Relevant"] = 10 - df.loc[i, "Irrelevant"]

    df = df.round(1).to_dict(orient="records")
    for d in df:
        d["x"] = d.pop("Total")

    df_relevant = [{k: v for k, v in d.items() if k != "Irrelevant"} for d in df]
    for d in df_relevant:
        d["y"] = d.pop("Relevant")

    df_irrelevant = [{k: v for k, v in d.items() if k != "Relevant"} for d in df]
    for d in df_irrelevant:
        d["y"] = d.pop("Irrelevant")

    payload = {"relevant": df_relevant, "irrelevant": df_irrelevant}

    return jsonify(payload)


@bp.route("/projects/<project_id>/progress_recall", methods=["GET"])
@login_required
@project_authorization
def api_get_progress_recall(project):
    """Get cumulative number of inclusions by ASReview/at random"""

    include_priors = request.args.get("priors", False, type=bool)

    with open_state(project.project_path) as s:
        if (
            project.config["reviews"][0]["status"] == "finished"
            and project.config["mode"] == PROJECT_MODE_SIMULATE
        ):
            data = _get_labels(s, priors=include_priors)
        else:
            data = s.get_labels(priors=include_priors)

        n_records = len(s.get_record_table())

    # create a dataset with the cumulative number of inclusions
    df = data.to_frame(name="Relevant").reset_index(drop=True).cumsum()
    df["Total"] = df.index + 1
    df["Random"] = (df["Total"] * (df["Relevant"][-1:] / n_records).values).round()

    df = df.round(1).to_dict(orient="records")
    for d in df:
        d["x"] = d.pop("Total")

    df_asreview = [{k: v for k, v in d.items() if k != "Random"} for d in df]
    for d in df_asreview:
        d["y"] = d.pop("Relevant")

    df_random = [{k: v for k, v in d.items() if k != "Relevant"} for d in df]
    for d in df_random:
        d["y"] = d.pop("Random")

    payload = {"asreview": df_asreview, "random": df_random}

    return jsonify(payload)


@bp.route("/projects/<project_id>/record/<doc_id>", methods=["POST", "PUT"])
@login_required
@project_authorization
def api_classify_instance(project, doc_id):  # noqa: F401
    """Label item

    This request handles the document identifier and the corresponding label.
    The result is stored in a temp location. If this storage exceeds a certain
    amount of values, then the model is triggered. The values of the location
    are passed to the model and the storaged is cleared. This model will run
    in the background.
    """
    # return the combination of document_id and label.
    record_id = int(request.form.get("doc_id"))
    label = int(request.form.get("label"))
    note = request.form.get("note", type=str)
    if not note:
        note = None

    is_prior = request.form.get("is_prior", default=False)

    retrain_model = False if is_prior == "1" else True
    prior = True if is_prior == "1" else False

    if request.method == "POST":
        with open_state(project.project_path, read_only=False) as state:
            # add the labels as prior data
            state.add_labeling_data(
                record_ids=[record_id], labels=[label], notes=[note], prior=prior
            )

    elif request.method == "PUT":
        with open_state(project.project_path, read_only=False) as state:
            if label in [0, 1]:
                state.update_decision(record_id, label, note=note)
            elif label == -1:
                state.delete_record_labeling_data(record_id)

    if retrain_model:
        # retrain model
        subprocess.Popen(
            [
                _get_executable(),
                "-m",
                "asreview",
                "web_run_model",
                str(project.project_path),
            ]
        )

    response = jsonify({"success": True})

    return response


@bp.route("/projects/<project_id>/get_document", methods=["GET"])
@login_required
@project_authorization
def api_get_document(project):  # noqa: F401
    """Retrieve documents in order of review.

    After these documents were retrieved, the queue on the client side is
    updated.
    This request can get triggered after each document classification.
    Although it might be better to call this function after 20 requests on the
    client side.
    """
    with open_state(project.project_path, read_only=False) as state:
        # First check if there is a pending record.
        _, _, pending = state.get_pool_labeled_pending()
        if not pending.empty:
            record_ids = pending.to_list()
        # Else query for a new record.
        else:
            record_ids = state.query_top_ranked(1)

    if len(record_ids) > 0:
        new_instance = record_ids[0]

        as_data = project.read_data()
        record = as_data.record(int(new_instance))

        item = {}
        item["title"] = record.title
        item["authors"] = record.authors
        item["abstract"] = record.abstract
        item["doi"] = record.doi
        item["url"] = record.url

        # return the debug label
        item["label_from_dataset"] = record.included

        item["doc_id"] = new_instance
        pool_empty = False
    else:
        # end of pool
        project.update_review(status="finished")
        item = None
        pool_empty = True

    return jsonify({"result": item, "pool_empty": pool_empty})


@bp.route("/projects/<project_id>/delete", methods=["DELETE"])
@login_required
@project_authorization
def api_delete_project(project):  # noqa: F401
    """"""

    if project.project_path.exists() and project.project_path.is_dir():
        try:
            # remove from database if applicable
            if not current_app.config.get("LOGIN_DISABLED", False):
                project = Project.query.filter(
                    and_(
                        Project.project_id == project.project_id,
                        Project.owner_id == current_user.id,
                    )
                ).one_or_none()

                if project is not None:
                    DB.session.delete(project)
                    DB.session.commit()
                else:
                    return jsonify(message="Failed to delete project in DB."), 500

            # and remove the folder
            shutil.rmtree(project.project_path)

        except Exception as err:
            logging.error(err)
            return jsonify(message="Failed to delete project."), 500

        response = jsonify({"success": True})

        return response



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["EmbeddingLSTM"]

import gzip
import io
import logging
import time
from multiprocessing import Process
from multiprocessing import Queue
from multiprocessing import cpu_count
from pathlib import Path
from urllib.request import urlopen

import numpy as np

from asreview.models.feature_extraction.base import BaseFeatureExtraction
from asreview.utils import get_data_home

try:
    from tensorflow.keras.preprocessing.sequence import pad_sequences
    from tensorflow.keras.preprocessing.text import Tokenizer
except ImportError:
    TF_AVAILABLE = False
else:
    TF_AVAILABLE = True


def _check_tensorflow():
    if not TF_AVAILABLE:
        raise ImportError("Install tensorflow package to use" " Embedding LSTM.")


class EmbeddingLSTM(BaseFeatureExtraction):
    """Embedding LSTM feature extraction technique (``embedding-lstm``).

    Feature extraction technique for
    :class:`asreview.models.classifiers.LSTMBaseClassifier` and
    :class:`asreview.models.classifiers.LSTMPoolClassifier` models.

    .. note::

        This feature extraction technique requires ``tensorflow`` to be
        installed. Use ``pip install asreview[tensorflow]`` or install all
        optional ASReview dependencies with ``pip install asreview[all]``

    Arguments
    ---------
    loop_sequence: bool
        Instead of zeros at the start/end of sequence loop it.
    num_words: int
        Maximum number of unique words to be processed.
    max_sequence_length: int
        Maximum length of the sequence. Shorter get struncated.
        Longer sequences get either padded with zeros or looped.
    padding: str
        Which side should be padded [pre/post].
    truncating:
        Which side should be truncated [pre/post].
    n_jobs:
        Number of processors used in reading the embedding matrix.
    """

    name = "embedding-lstm"
    label = "Embedding LSTM"

    def __init__(
        self,
        *args,
        loop_sequence=1,
        num_words=20000,
        max_sequence_length=1000,
        padding="post",
        truncating="post",
        n_jobs=1,
        **kwargs,
    ):
        """Initialize the embedding matrix feature extraction."""
        super(EmbeddingLSTM, self).__init__(*args, **kwargs)
        self.embedding = None
        self.num_words = num_words
        self.max_sequence_length = max_sequence_length
        self.padding = padding
        self.truncating = truncating
        self.n_jobs = n_jobs
        self.loop_sequence = loop_sequence

    def transform(self, texts):
        _check_tensorflow()

        self.X, self.word_index = text_to_features(
            texts,
            loop_sequence=self.loop_sequence,
            num_words=self.num_words,
            max_sequence_length=self.max_sequence_length,
            padding=self.padding,
            truncating=self.truncating,
        )
        return self.X

    def get_embedding_matrix(self, texts, embedding_fp):
        _check_tensorflow()

        self.fit_transform(texts)
        if embedding_fp is None:
            embedding_fp = Path(get_data_home(), EMBEDDING_EN["name"]).expanduser()

            if not embedding_fp.exists():
                logging.warning(
                    "Warning: will start to download large "
                    "embedding file in 10 seconds."
                )
                time.sleep(10)
                download_embedding()
        logging.info("Loading embedding matrix. " "This can take several minutes.")

        embedding = load_embedding(embedding_fp, n_jobs=self.n_jobs)
        return sample_embedding(embedding, self.word_index)

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_space, hyper_choices = super(EmbeddingLSTM, self).full_hyper_space()
        hyper_space.update({"fex_loop_sequences": hp.randint("fex_loop_sequences", 2)})
        return hyper_space, hyper_choices


EMBEDDING_EN = {
    "url": "https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.en.300.vec.gz",  # noqa
    "name": "fasttext.cc.en.300.vec",
}


def loop_sequences(X, max_sequence_length=1000):
    # Loop the sequences instead of padding.
    for i, old_x in enumerate(X):
        nz = max_sequence_length - 1
        while nz >= 0 and old_x[nz] == 0:
            nz -= 1
        # If there are only 0's (no data), continue.
        if nz < 0:
            continue
        nz += 1
        new_x = old_x.copy()

        j = 1
        # Copy the old data to the new matrix.
        while nz * j < max_sequence_length:
            cp_len = min(nz * (j + 1), max_sequence_length) - nz * j
            new_x[nz * j : nz * j + cp_len] = old_x[0:cp_len]
            j += 1
        X[i] = new_x
    return X


def text_to_features(
    sequences,
    loop_sequence=1,
    num_words=20000,
    max_sequence_length=1000,
    padding="post",
    truncating="post",
):
    """Convert text data into features.

    Arguments
    ---------
    sequences: list, numpy.ndarray, pandas.Series
        The sequences to convert into features.
    num_words: int
        See keras Tokenizer

    Returns
    -------
    np.ndarray, dict
        The array with features and the dictiory that maps words to values.
    """

    # fit on texts
    tokenizer = Tokenizer(num_words=num_words)
    tokenizer.fit_on_texts(sequences)

    # tokenize sequences
    tokens = tokenizer.texts_to_sequences(sequences)

    # Pad sequences with zeros.
    x = pad_sequences(
        tokens, maxlen=max_sequence_length, padding=padding, truncating=truncating
    )

    if loop_sequence == 1:
        x = loop_sequences(x, max_sequence_length)
    # word index hack. see issue
    # https://github.com/keras-team/keras/issues/8092
    word_index = {e: i for e, i in tokenizer.word_index.items() if i <= num_words}

    return x, word_index


def _embedding_reader(filename, input_queue, block_size=1000):
    """Process that reads the word embeddings from a file.

    Parameters
    ----------
    filename: str
        File of trained embedding vectors.
    input_queue: Queue
        Queue to store jobs in.
    block_size: int
        Number of lines for each job.
    """

    with open(filename, "r", encoding="utf-8", newline="\n") as f:
        # Throw away the first line, since we don't care about the dimensions.
        f.readline()

        i_line = 0
        buffer = []
        # Read the embedding file line by line.
        for line in f:
            i_line += 1
            buffer.append(line)
            # If the buffer is full, write it to the queue.
            if i_line == block_size:
                input_queue.put(buffer)
                i_line = 0
                buffer = []
        if i_line > 0:
            input_queue.put(buffer)

    # Put the string "DONE" in the queue, to ensure that the
    # worker processes finish.

    input_queue.put("DONE")


def _embedding_worker(input_queue, output_queue, emb_vec_dim, word_index=None):
    """Process that reads the word embeddings from a file.

    Parameters
    ----------
    input_queue: Queue
        Queue in which the jobs are submitted.
    output_queue: Queue
        Queue to store the embedding in dictionary form.
    emb_vec_dim: int
        Dimension of each embedding vector.
    word_index: dict
        Dictionary of the sample embedding.
    """

    bad_input = False
    bad_values = {}
    while True:
        embedding = {}
        buffer = input_queue.get()
        if buffer == "DONE":
            break

        for line in buffer:
            line = line.rstrip()
            values = line.split(" ")

            if len(values) != emb_vec_dim + 1:
                if not bad_input:
                    print("Error: bad input in embedding vector.")
                bad_input = True
                bad_values = values
                break

            word = values[0]
            if word_index is not None and word not in word_index:
                continue
            coefs = values[1 : emb_vec_dim + 1]

            # store the results
            embedding[word] = np.asarray(coefs, dtype=np.float32)
        output_queue.put(embedding)

    # We removed the "DONE" from the input queue, so put it back in for
    # the other processes.
    input_queue.put("DONE")

    # Store the results in the output queue
    if bad_input:
        output_queue.put({"ErrorBadInputValues": bad_values})
    output_queue.put("DONE")


def _embedding_aggregator(output_queue, n_worker):
    """Process that aggregates the results of the workers.
        This should be the main/original process.

    Parameters
    ----------
    output_queue: Queue
        This queue is the output queue of the workers.
    n_worker: int
        The number of worker processes.

    Returns
    -------
    Aggregated embedding dictionary.
    """

    embedding = {}

    num_done = 0
    while num_done < n_worker:
        new_embedding = output_queue.get()
        if new_embedding == "DONE":
            num_done += 1
        else:
            embedding.update(new_embedding)

    return embedding


def download_embedding(
    url=EMBEDDING_EN["url"], name=EMBEDDING_EN["name"], data_home=None
):
    """Download word embedding file.

    Download word embedding file, unzip the file and save to the
    file system.

    Parameters
    ----------
    url: str
        The URL of the gzipped word embedding file
    name: str
        The filename of the embedding file.
    data_home: str
        The location of the ASR datasets.
        Default `asreview.utils.get_data_home()`

    """

    if data_home is None:
        data_home = get_data_home()

    out_fp = Path(data_home, name)

    logging.info(f"Start downloading: {url}")

    r = urlopen(url)
    compressed_file = io.BytesIO(r.read())

    logging.info(f"Save embedding to {out_fp}")

    decompressed_file = gzip.GzipFile(fileobj=compressed_file)

    with open(out_fp, "wb") as out_file:
        for line in decompressed_file:
            out_file.write(line)


def load_embedding(fp, word_index=None, n_jobs=None):
    """Load embedding matrix from file.

    The embedding matrix needs to be stored in the
    FastText format.

    Parameters
    ----------
    fp: str
        File path of the trained embedding vectors.
    word_index: dict
        Sample word embeddings.
    n_jobs: int
        Number of processes to parse the embedding (+1 process for reading).
    verbose: int
        The verbosity. Default 1.


    Returns
    -------
    dict:
        The embedding weights stored in a dict with the word as key and
        the weights as values.
    """

    # Maximum number of jobs in the queue.
    queue_size = 500

    # Set the number of reader processes to use.
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs == -1:
        n_jobs = cpu_count() - 1

    input_queue = Queue(queue_size)
    output_queue = Queue()

    with open(fp, "r", encoding="utf-8", newline="\n") as f:
        n_words, emb_vec_dim = list(map(int, f.readline().split(" ")))

    logging.debug(f"Reading {n_words} vectors with {emb_vec_dim} dimensions.")

    worker_procs = []
    p = Process(target=_embedding_reader, args=(fp, input_queue), daemon=True)
    worker_procs.append(p)
    for _ in range(n_jobs):
        p = Process(
            target=_embedding_worker,
            args=(input_queue, output_queue, emb_vec_dim, word_index),
            daemon=True,
        )
        worker_procs.append(p)

    # Start workers.
    for proc in worker_procs:
        proc.start()
    embedding = _embedding_aggregator(output_queue, n_jobs)

    # Merge dictionaries of workers

    # Join workers
    for proc in worker_procs:
        proc.join()

    if "ErrorBadInputValues" in embedding:
        bad_values = embedding["ErrorBadInputValues"]
        raise ValueError(f"Check embedding matrix, bad format: {bad_values}")

    logging.debug(f"Found {len(embedding)} word vectors.")

    return embedding


def sample_embedding(embedding, word_index):
    """Sample embedding matrix

    Parameters
    ----------
    embedding: dict
        A dictionary with the words and embedding vectors.
    word_index: dict
        A word_index like the output of Keras Tokenizer.word_index.
    verbose: int
        The verbosity. Default 1.

    Returns
    -------
    (np.ndarray, list):
        The embedding weights strored in a two dimensional
        numpy array and a list with the corresponding words.
    """

    n_words, emb_vec_dim = len(word_index), len(next(iter(embedding.values())))

    logging.debug(
        f"Creating matrix with {n_words} vectors " f"with dimension {emb_vec_dim}."
    )

    # n+1 because 0 is preserved in the tokenizing process.
    embedding_matrix = np.zeros((n_words + 1, emb_vec_dim))

    for word, i in word_index.items():
        coefs = embedding.get(word)
        if coefs is not None:
            embedding_matrix[i] = coefs
    logging.debug(f"Shape of embedding matrix: {embedding_matrix.shape}")

    return embedding_matrix



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = []

import json
import shutil
import sqlite3
import time
from base64 import b64decode
from datetime import datetime
from io import BytesIO
from pathlib import Path
from uuid import uuid4

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse import issparse
from scipy.sparse import load_npz
from scipy.sparse import save_npz

from asreview._version import get_versions
from asreview.state.errors import StateError
from asreview.state.legacy.utils import open_state as open_state_legacy

SQLSTATE_VERSION = "1.0"
ASREVIEW_FILE_EXTENSION = ".asreview"


class StateConversionException(Exception):
    pass


def is_old_project(fp):
    """Check if state file is old version."""
    if Path(fp, "reviews").is_dir():
        return False
    else:
        return True


def get_old_project_status(config):
    # project is marked as finished
    if config.get("reviewFinished", False):
        return "finished"

    # project init is not ready
    if "projectInitReady" in config and not config["projectInitReady"]:
        return "setup"

    # project init flag is not available
    if "projectInitReady" not in config:
        if "projectHasPriorKnowledge" in config:
            if config["projectHasPriorKnowledge"]:
                return "review"
            else:
                return "setup"

    return "review"


def decode_feature_matrix(jsonstate, data_hash):
    """Get the feature matrix from a json state as a scipy csr_matrix."""
    my_data = jsonstate._state_dict["data_properties"][data_hash]
    encoded_X = my_data["feature_matrix"]
    matrix_type = my_data["matrix_type"]
    if matrix_type == "ndarray":
        return csr_matrix(encoded_X)
    elif matrix_type == "csr_matrix":
        with BytesIO(b64decode(encoded_X)) as f:
            return load_npz(f)
    return encoded_X


def upgrade_asreview_project_file(fp, from_version=0, to_version=1):
    """Convert an old asreview project folder to the new format.

    Arguments
    ---------
    fp: str/path
        Location of the (unzipped) project file.

    Returns
    -------
    Converts the data in the project to the new format
    and adds it to the folder in place."""

    if from_version != 0 and to_version != 1:
        raise ValueError(
            f"Not possible to upgrade from v{from_version} to v{to_version}."
        )

    # Check if it is indeed an old format project.
    if not is_old_project(fp):
        raise ValueError(
            f"There already is a 'reviews' folder at {fp}. "
            f"This project seems to be in new format."
        )

    # Current Paths
    fp = Path(fp)
    legacy_fp = Path(fp, "legacy")
    move_old_files_to_legacy_folder(fp)

    try:
        # Current paths.
        json_fp = Path(legacy_fp, "result.json")
        labeled_json_fp = Path(legacy_fp, "labeled.json")
        pool_fp = Path(legacy_fp, "pool.json")
        kwargs_fp = Path(legacy_fp, "kwargs.json")
        review_id = str(uuid4().hex)

        # Create the reviews folder and the paths for the results and settings.
        Path(fp, "reviews", review_id).mkdir(parents=True)
        sql_fp = str(Path(fp, "reviews", review_id, "results.sql"))
        settings_metadata_fp = Path(fp, "reviews", review_id, "settings_metadata.json")

        # Create the path for the feature matrix.

        # Create sqlite table with the results of the review.
        convert_json_results_to_sql(sql_fp, json_fp, labeled_json_fp)

        # Create sqlite tables 'last_probabilities'.
        convert_json_last_probabilities(sql_fp, json_fp)

        # Create the table for the last ranking of the model.
        create_last_ranking_table(sql_fp, pool_fp, kwargs_fp, json_fp)

        # Add the record table to the sqlite database as the table 'record_table'.
        convert_json_record_table(sql_fp, json_fp)

        # Create decision changes table.
        create_decision_changes_table(sql_fp)

        # Create json for settings.
        convert_json_settings_metadata(settings_metadata_fp, json_fp)

        # Create file for the feature matrix.
        with open(kwargs_fp, "r") as f:
            kwargs_dict = json.load(f)
            feature_extraction_method = kwargs_dict["feature_extraction"]
        feature_matrix_fp = convert_json_feature_matrix(
            fp, json_fp, feature_extraction_method
        )

        # --- Upgrade the project.json file.

        # extract the start time from the state json
        with open(json_fp, "r") as f:
            start_time = json.load(f)["time"]["start_time"]
            start_time = datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S.%f")

        # open the project json and upgrade
        with open(Path(fp, "project.json"), "r") as f:
            project_config_old = json.load(f)

        project_config_new = upgrade_project_config(
            project_config_old,
            review_id,
            start_time,
            Path(feature_matrix_fp).name,
            feature_extraction_method,
        )

        # dump the project json
        with open(Path(fp, "project.json"), "w") as f:
            json.dump(project_config_new, f)
    except Exception as e:
        rollback_conversion(fp, check_is_converted=False)
        raise StateConversionException(
            f"An error occurred during conversion of state "
            f"from {from_version} to {to_version}."
        ) from e


def move_old_files_to_legacy_folder(fp):
    """Move the old files to a legacy folder.

    Arguments
    ----------
    fp: pathlib.Path
        Location of the (unzipped) project file.

    Returns
    -------
    Creates a folder 'legacy' in the project file, moves all current files to
    this legacy folder, and keeps a copy of 'project.json' and the data folder
    at the original place.
    """

    project_content = list(fp.iterdir())

    # copy to legacy folder
    shutil.copytree(fp, Path(fp, "legacy"))

    # remove files and folders
    files_to_keep = ["project.json", "data", "lock.sqlite"]

    for f in project_content:
        if f.name not in files_to_keep:
            if f.is_file():
                f.unlink()
            elif f.is_dir():
                shutil.rmtree(f)
            else:
                pass


def upgrade_project_config(
    config,
    review_id=None,
    start_time=None,
    feature_matrix_name=None,
    feature_extraction_method=None,
):
    """Update the project.json file to contain the review information , the
    feature matrix information and the new state version number.

    Arguments
    ---------
    config: str/path
        Path to the project json file.
    review_id: str
        Identifier of the review.
    start_time: str
        String containing start time of the review.
    feature_matrix_fp: str/path
        Location of the feature matrix.
    feature_extraction_method: str
        Name of the feature extraction method.
    """

    start_time_s = str(start_time) if start_time else None

    # Add the review information.
    config["reviews"] = [
        {
            "id": review_id,
            "start_time": start_time_s,
            "status": get_old_project_status(config),
        }
    ]

    # Add the feature matrix information.
    config["feature_matrices"] = [
        {"id": feature_extraction_method, "filename": feature_matrix_name}
    ]

    # Add the project mode.
    config["mode"] = config.get("mode", "oracle")

    # Update the state version.
    config["version"] = get_versions()["version"]
    config["state_version"] = SQLSTATE_VERSION

    # set created_at_unix to start time (empty: None)
    if "created_at_unix" not in config:
        try:
            config["created_at_unix"] = time.mktime(start_time.timetuple())
        except Exception:
            config["created_at_unix"] = None

    config["datetimeCreated"] = start_time_s

    # delete deprecated metadata
    config.pop("projectInitReady", None)
    config.pop("projectHasPriorKnowledge", None)
    config.pop("projectHasDataset", None)

    return config


def convert_json_settings_metadata(fp, json_fp):
    """Get the settings and metadata from a json state and save it as
    a json file at the location given by fp.

    Arguments
    ---------
    fp: str/path
        Path where to save the json file.
    json_fp: str/path
        Path to the json state file.
    """
    data_dict = {}
    with open_state_legacy(json_fp) as json_state:
        data_dict["settings"] = json_state._state_dict["settings"]
        # The 'triple' balance strategy is no longer implemented.
        if data_dict["settings"]["balance_strategy"] == "triple":
            data_dict["settings"]["balance_strategy"] = "double"
        data_dict["state_version"] = SQLSTATE_VERSION
        data_dict["software_version"] = json_state._state_dict["software_version"]
        data_dict["model_has_trained"] = True

        # remove the outdated mode
        data_dict.pop("mode", None)

    with open(fp, "w") as f:
        json.dump(data_dict, f)


def create_last_ranking_table(sql_fp, pool_fp, kwargs_fp, json_fp):
    """Create the table which will contain the ranking of the last iteration of
    the model.

    Arguments
    ---------
    sql_fp: str/path
        Path where to save the record table. Should be a .sql file.
    """

    with open(pool_fp) as f_pool:
        pool_ranking = json.load(f_pool)

    with open(kwargs_fp, "r") as f_kwargs:
        kwargs_dict = json.load(f_kwargs)

    # Add the record_ids not found in the pool to the end of the ranking.
    with open_state_legacy(json_fp) as json_state:
        record_table = get_json_record_table(json_state)
    records_not_in_pool = [
        record_id for record_id in record_table if record_id not in pool_ranking
    ]
    pool_ranking += records_not_in_pool

    # Convert the records in the pool to the new record ids (starting from 0).
    old_to_new_record_ids = {old_id: idx for idx, old_id in enumerate(record_table)}
    pool_ranking = [old_to_new_record_ids[record] for record in pool_ranking]

    # Set the training set to -1 (prior) for records from old pool.
    training_set = -1
    time = None

    last_ranking = [
        (
            v,
            i,
            kwargs_dict["model"],
            kwargs_dict["query_strategy"],
            kwargs_dict["balance_strategy"],
            kwargs_dict["feature_extraction"],
            training_set,
            time,
        )
        for i, v in enumerate(pool_ranking)
    ]

    with sqlite3.connect(sql_fp) as con:
        cur = con.cursor()

        # Create the last_ranking table.
        cur.execute(
            """CREATE TABLE last_ranking
                        (record_id INTEGER,
                        ranking INT,
                        classifier TEXT,
                        query_strategy TEXT,
                        balance_strategy TEXT,
                        feature_extraction TEXT,
                        training_set INTEGER,
                        time INTEGER)"""
        )
        cur.executemany(
            """INSERT INTO last_ranking VALUES
                                    (?, ?, ?, ?, ?, ?, ?, ?)""",
            last_ranking,
        )
        con.commit()


def convert_json_last_probabilities(sql_fp, json_fp):
    """Get the last ranking from a json state and save it as the table
    'last_probabilities' in the .sql file at the location of sql_fp.

    Arguments
    ---------
    sql_fp: str/path
        Path where to save the record table. Should be a .sql file.
    json_fp: str/path
        Path to the json state file.
    """
    with open_state_legacy(json_fp) as json_state:
        # Get the last predicted probabilities from the state file.
        # Also get the number of record labeled and the classifier.
        last_probabilities = json_state.pred_proba

        # Put them in the format for input in the sqlite database.
        last_probabilities = [(proba,) for proba in last_probabilities]

        with sqlite3.connect(sql_fp) as con:
            cur = con.cursor()
            cur.execute(
                """CREATE TABLE last_probabilities
                            (proba REAL)"""
            )
            cur.executemany(
                """INSERT INTO last_probabilities VALUES
                                        (?)""",
                last_probabilities,
            )

            con.commit()


def get_json_state_data_hash(json_state):
    """Get the data hash from a json state."""
    return list(json_state._state_dict["data_properties"].keys())[0]


def get_json_record_table(json_state):
    """Get the record table from a json state."""
    data_hash = get_json_state_data_hash(json_state)
    record_table = json_state._state_dict["data_properties"][data_hash]["record_table"]
    return record_table


def convert_json_feature_matrix(fp, json_fp, feature_extraction_method):
    """Get the feature matrix from a json state file. Save it in the feature
    matrices folder. Format is .npz if the matrix is sparse and .npy if the
    matrix is dense.

    Arguments
    ---------
    fp: str/path
        Project folder.
    json_fp: str/path
        Path to the json state file.
    feature_extraction_method: str
        Name of the feature extraction method.

    Returns
    -------
    pathlib.Path
        Path where the feature matrix is saved.
    """
    feature_matrices_fp = Path(fp, "feature_matrices")
    feature_matrices_fp.mkdir()

    with open_state_legacy(json_fp) as json_state:
        data_hash = get_json_state_data_hash(json_state)
        feature_matrix = decode_feature_matrix(json_state, data_hash)
        if issparse(feature_matrix):
            save_fp = Path(
                feature_matrices_fp, f"{feature_extraction_method}_feature_matrix.npz"
            )
            save_npz(save_fp, feature_matrix)
        else:
            save_fp = Path(
                feature_matrices_fp, f"{feature_extraction_method}_feature_matrix.npy"
            )
            np.save(save_fp, feature_matrix)

    return save_fp


def convert_json_record_table(sql_fp, json_fp):
    """Get the record table and save as the table 'record_table'
    in the .sql file at results_fp.

    Arguments
    ---------
    sql_fp: str/path
        Path where to save the record table. Should be a .sql file.
    json_fp: str/path
        Path to the json state file.
    """

    with open_state_legacy(json_fp) as json_state:
        record_table = get_json_record_table(json_state)

    # Convert record_table to list of tuples.
    record_table = [(record_id,) for record_id in range(len(record_table))]

    con = sqlite3.connect(sql_fp)
    cur = con.cursor()
    cur.execute(
        """CREATE TABLE record_table
                    (record_id INT)"""
    )
    cur.executemany(
        """INSERT INTO record_table VALUES
                                (?)""",
        record_table,
    )
    con.commit()
    con.close()


def convert_json_results_to_sql(sql_fp, json_fp, labeled_json_fp):
    """Convert the result of a json state file to a sqlite database."""
    with open_state_legacy(json_fp, read_only=True) as sf:
        with sqlite3.connect(sql_fp) as con:
            with open(labeled_json_fp, "r") as file:
                labeled_json = json.load(file)

            cur = con.cursor()

            # Create the results table.
            cur.execute(
                """CREATE TABLE results
                            (record_id INTEGER,
                            label INTEGER,
                            classifier TEXT,
                            query_strategy TEXT,
                            balance_strategy TEXT,
                            feature_extraction TEXT,
                            training_set INTEGER,
                            labeling_time INTEGER,
                            notes TEXT)"""
            )

            record_table = get_json_record_table(sf)
            record_id_to_row_number = {
                record_table[i]: i for i in range(len(record_table))
            }
            old_record_ids = [x[0] for x in labeled_json]
            sf_indices = [
                record_id_to_row_number[record_id] for record_id in old_record_ids
            ]

            sf_labels = [x[1] for x in labeled_json]

            # query strategy.
            old_query_strategy = [
                sample_data[2]
                for query in range(len(sf._state_dict["results"]))
                for sample_data in sf._state_dict["results"][query]["labelled"]
            ]

            n_priors = old_query_strategy.count("prior")
            n_records_labeled = len(sf_indices)
            n_non_prior_records = n_records_labeled - n_priors

            query_strategy = sf.settings.to_dict()["query_strategy"]
            sf_query_strategy = ["prior"] * n_priors + [
                query_strategy
            ] * n_non_prior_records

            # classifier.
            classifier = sf.settings.to_dict()["model"]
            sf_classifiers = [None] * n_priors + [
                f"{classifier}" for _ in range(n_non_prior_records)
            ]

            # training set.
            sf_training_sets = [-1] * n_priors + list(
                range(n_priors, n_records_labeled)
            )

            # feature extraction.
            feature_extraction = sf.settings.to_dict()["feature_extraction"]
            sf_feature_extraction = [None] * n_priors + [
                f"{feature_extraction}" for _ in range(n_non_prior_records)
            ]

            # balance strategy.
            balance_strategy = sf.settings.to_dict()["balance_strategy"]
            sf_balance_strategy = [None] * n_priors + [
                f"{balance_strategy}" for _ in range(n_non_prior_records)
            ]

            # Labeling time.
            sf_time = [0 for _ in range(n_records_labeled)]

            # No notes were saved before.
            sf_notes = [None for _ in range(n_records_labeled)]

            # Check that all datasets have the same number of entries.
            lengths = [
                len(sf_indices),
                len(sf_labels),
                len(sf_classifiers),
                len(sf_training_sets),
                len(sf_query_strategy),
                len(sf_time),
                len(sf_feature_extraction),
                len(sf_balance_strategy),
                len(sf_notes),
            ]
            if not all([length == n_records_labeled for length in lengths]):
                raise StateError("All datasets should have the same number of entries.")

            # Create the database rows.
            db_rows = [
                (
                    sf_indices[i],
                    sf_labels[i],
                    sf_classifiers[i],
                    sf_query_strategy[i],
                    sf_balance_strategy[i],
                    sf_feature_extraction[i],
                    sf_training_sets[i],
                    sf_time[i],
                    sf_notes[i],
                )
                for i in range(n_records_labeled)
            ]
            cur.executemany(
                """INSERT INTO results VALUES
                            (?, ?, ?, ?, ?, ?, ?, ?, ?)""",
                db_rows,
            )
            con.commit()


def create_decision_changes_table(sql_fp):
    """Create an emtpy table that will contain the record_ids and new labels
    of the records whose label was changed after the original labeling action.
    Also contains the time at which the label was changed."""
    with sqlite3.connect(sql_fp) as con:
        cur = con.cursor()

        cur.execute(
            """CREATE TABLE decision_changes
                                    (record_id INTEGER,
                                    new_label INTEGER,
                                    time INTEGER)"""
        )

        con.commit()


# Disable is_converted_check.
def rollback_conversion(fp, from_version=1, to_version=0, check_is_converted=True):
    if from_version != 1 and to_version != 0:
        raise ValueError(
            f"Not possible to roll back conversion from v{from_version} "
            f"to v{to_version}."
        )

    if check_is_converted and not is_converted_project(fp):
        raise ValueError(f"Project file at {fp} is not a converted " f"project file.")

    fp = Path(fp)
    legacy_fp = Path(fp, "legacy")

    # Delete everything other than the legacy folder.
    for item in fp.iterdir():
        if item.is_file():
            item.unlink()
        elif item.is_dir() and (item.name != "legacy"):
            shutil.rmtree(item)
        else:
            pass

    # Copy from legacy folder and delete it after.
    shutil.copytree(legacy_fp, fp, dirs_exist_ok=True)
    shutil.rmtree(legacy_fp)


def is_converted_project(fp):
    """Check if asreview file has been converter from v0 to v1."""
    # Check if there is a legacy project.json and it has v0.
    try:
        with open(Path(fp, "legacy", "project.json"), "r") as f:
            project_config_old = json.load(f)
    except FileNotFoundError:
        return False

    if project_config_old["version"][0] != "0":
        return False

    # Check if the current project.json has 'state_version' == 1.
    with open(Path(fp, "project.json"), "r") as f:
        project_config_current = json.load(f)

    try:
        current_state_version = project_config_current["state_version"]
    except KeyError:
        return False

    if current_state_version[0] == "1":
        return True
    else:
        return False
